from __future__ import annotations

from typing import TYPE_CHECKING, Any, List, Type, TypeVar

from vectice.api.http_error_handlers import ClientErrorHandler
from vectice.api.json import (
    CodeOutput,
    CodeVersionOutput,
    DatasetRegisterOutput,
    IterationOutput,
    ModelRegisterOutput,
    PagedResponse,
    PhaseOutput,
    ProjectOutput,
    PublicConfigOutput,
    StepOutput,
    UserActivity,
    UserAndDefaultWorkspaceOutput,
    WorkspaceOutput,
)

if TYPE_CHECKING:
    from gql import Client

    from vectice.api._auth import Auth

ResultType = TypeVar("ResultType")


class Parser:
    def map_type(self, type_name: str) -> Type[ResultType]:
        types_map = {
            "Workspace": WorkspaceOutput,
            "Project": ProjectOutput,
            "Phase": PhaseOutput,
            "IterationStep": StepOutput,
            "Iteration": IterationOutput,
            "DatasetRegisterResultOutput": DatasetRegisterOutput,
            "ModelRegisterResultOutput": ModelRegisterOutput,
            "UserActivity": UserActivity,
            "Code": CodeOutput,
            "CodeVersion": CodeVersionOutput,
            "PublicConfigOutput": PublicConfigOutput,
            "UserAndDefaultWorkspaceOutput": UserAndDefaultWorkspaceOutput,
            "StepDefinition": StepOutput,
        }
        if type_name in types_map.keys():
            return types_map[type_name]  # type: ignore[return-value]
        else:
            raise RuntimeError("Unknown type: " + type_name)

    def parse_item(self, item: dict[str, Any]) -> ResultType:  # type: ignore[type-var]
        clazz: Type[ResultType] = self.map_type(item["__typename"])
        result = clazz(**item)
        return result

    def parse_list(self, list: list[dict[str, Any]]) -> list[ResultType]:
        # TODO: rename 'list' parameter to avoid shadowing builtin
        result: List[ResultType] = []
        for item in list:
            result.append(self.parse_item(item))
        return result

    def parse_paged_response(self, data: dict[str, Any]) -> PagedResponse[ResultType]:
        return PagedResponse(data["total"], data["page"], self.map_type(data["items"][0]["__typename"]), data["items"])

    def parse(self, data) -> ResultType | list[ResultType] | PagedResponse[ResultType]:
        if type(data) == list:
            return self.parse_list(data)
        else:
            if "items" in data and "page" in data and "total" in data:
                return self.parse_paged_response(data)
            else:
                return self.parse_item(data)


class GqlApi:
    def __init__(self, client: Client, auth: Auth):
        self.client = client
        self.auth = auth
        self._error_handler = ClientErrorHandler()

    def execute(self, query, variables=None):
        # retrieve auth & api http headers
        self.client.transport.headers = self.auth.http_headers
        return self.client.execute(query, variables)

    @staticmethod
    def build_query(
        gql_query: str,
        variable_types: str | None = None,
        returns: str | None = None,
        keyword_arguments: str | None = None,
        query: bool = True,
    ):
        query_type = "query" if query else "mutation"
        variable_types_formatted = f"({variable_types})" if variable_types else ""
        keyword_arguments_formatted = f"({keyword_arguments})" if keyword_arguments else ""
        gql_returns = f"{{{returns}}}" if returns else ""
        query_built = f"{query_type} {gql_query} {variable_types_formatted} {{{gql_query} {keyword_arguments_formatted} {gql_returns} }}"
        return query_built



from __future__ import annotations

import datetime
import glob
import hashlib
import logging
import os

from vectice.models.datasource.datawrapper.data_wrapper import DataWrapper
from vectice.models.resource.metadata import (
    DatasetSourceOrigin,
    DatasetSourceUsage,
    DatasetType,
    File,
    FilesMetadata,
)
from vectice.utils.deprecation import deprecate

_logger = logging.getLogger(__name__)


class FileDataWrapper(DataWrapper):
    """Deprecated. Wrap columnar data and its metadata in a local file.

    WARNING: **Wrappers are deprecated.**
    Instead, use [`Dataset`][vectice.models.dataset.Dataset]
    and [`FileResource`][vectice.models.resource.file_resource.FileResource].

    Vectice stores metadata -- data about your dataset -- communicated
    with a DataWrapper.  Your actual dataset is not stored by Vectice.

    This DataWrapper wraps data that you have stored in a local file.

    ```python
    from vectice import DatasetType, FileDataWrapper, connect

    my_project = connect(...)  # (1)
    my_phase = my_project.phase(...)  # (2)
    my_iter = my_phase.iteration()  # (3)

    my_iter.step_my_data = FileDataWrapper(
        path="my/file/path",
        name="My origin dataset name",
        type=DatasetType.ORIGIN,
    )
    ```

    1. See [connection][vectice.Connection.connect].
    1. See [phases][vectice.models.Phase].
    1. See [iterations][vectice.models.Iteration].

    Note that these three concepts are distinct, even if easily conflated:

    * Where the data is stored
    * The format at rest (in storage)
    * The format when loaded in a running python program

    Notably, the statistics collectors provided by Vectice operate
    only on this last and only in the even that the data is loaded as
    a pandas dataframe.
    """

    @deprecate(
        parameter="inputs",
        warn_at="23.1",
        fail_at="23.2",
        remove_at="23.3",
        reason="The 'inputs' parameter is renamed 'derived_from'. "
        "Using 'inputs' will raise an error in v{fail_at}. "
        "The parameter will be removed in v{remove_at}.",
    )
    def __init__(
        self,
        path: str,
        name: str,
        usage: DatasetSourceUsage | None = None,
        derived_from: list[int] | None = None,
        inputs: list[int] | None = None,
        type: DatasetType | None = None,
    ):
        """Initialize a file data wrapper.

        Parameters:
            path: The path of the file to wrap.
            name: The name of the DataWrapper (local to Vectice).
            usage: The usage of the dataset.
            derived_from: The list of dataset ids to derive this new dataset from.
            inputs: Deprecated. Use `derived_from` instead.
            type: The type of the dataset.

        Examples:
            The following example shows how to wrap a CSV file
            called `iris.csv` in the current directory:

            >>> from vectice import FileDataWrapper, DatasetSourceUsage, DatasetType
            >>> iris_trainset = FileDataWrapper(
            ...     path="iris.csv",
            ...     name="training dataset",
            ...     type=DatasetType.MODELING,
            ...     usage=DatasetSourceUsage.TRAINING,
            ...     derived_from=[64, 128],
            ... )
            2023/02/01 15:43:59 INFO vectice.models.datasource.datawrapper.file_data_wrapper: File: iris.csv wrapped successfully.
        """
        if not derived_from and inputs:
            derived_from = inputs

        self.path = path
        super().__init__(name=name, type=type, usage=usage, derived_from=derived_from)
        _logger.info(f"File: {path} wrapped successfully.")

    def _fetch_data(self) -> dict[str, bytes]:
        datas = {}
        for file in self.metadata.files:
            with open(file.name, "rb") as opened_file:
                datas[file.name] = opened_file.read()
        return datas

    def _build_metadata(self, path: str | None = None) -> FilesMetadata:
        if self.path is None:
            raise ValueError("Path can not be None.")
        files, total_size = self._file_visitor_list_builder(self.path, [])
        metadata = FilesMetadata(size=total_size, origin=DatasetSourceOrigin.LOCAL.value, files=files, usage=self.usage)
        return metadata

    def _file_visitor_list_builder(self, path: str, files: list[File]) -> tuple[list[File], int]:
        total_size = 0
        if not os.path.exists(path):
            raise FileNotFoundError
        if os.path.isfile(path):
            file: File = self._build_file_from_path(path)
            total_size += file.size
            files.append(file)
        else:
            for entry_path in glob.glob(f"{path}/**", recursive=True):
                if os.path.isfile(entry_path):
                    glob_file: File = self._build_file_from_path(entry_path)
                    total_size += glob_file.size
                    files.append(glob_file)
        return files, total_size

    def _build_file_from_path(self, path: str) -> File:
        entry_stats: os.stat_result = os.stat(path)
        name = path.split("\\")[-1]
        return File(
            name=name,
            size=entry_stats.st_size,
            fingerprint=self._compute_digest_for_path(path),
            updated_date=self._convert_date_to_iso(entry_stats.st_mtime),
            created_date=self._convert_date_to_iso(self._get_created_date(entry_stats)),
            uri=f"file://{os.path.abspath(path)}",
        )

    @staticmethod
    def _get_created_date(entry_stats: os.stat_result) -> float:
        try:
            return float(entry_stats.st_birthtime)  # type: ignore[attr-defined]
        except AttributeError:
            return entry_stats.st_ctime

    @staticmethod
    def _compute_digest_for_path(path: str) -> str:
        sha = hashlib.sha256()
        bytes_array = bytearray(128 * 1024)
        mv = memoryview(bytes_array)
        with open(path, "rb", buffering=0) as file:
            n = file.readinto(mv)
            while n:
                sha.update(mv[:n])
                n = file.readinto(mv)
        return sha.hexdigest()

    @staticmethod
    def _convert_date_to_iso(timestamp: float) -> str:
        return datetime.datetime.fromtimestamp(timestamp).isoformat()



from __future__ import annotations

from vectice.models.resource.metadata.base import (
    DatasetSourceType,
    DatasetSourceUsage,
    Metadata,
)


class FilesMetadata(Metadata):
    """The metadata of a set of files."""

    def __init__(
        self,
        files: list[File],
        size: int,
        usage: DatasetSourceUsage | None = None,
        origin: str | None = None,
    ):
        """Initialize a FilesMetadata instance.

        Parameters:
            files: The list of files of the dataset.
            size: The size of the set of files.
            usage: The usage of the dataset.
            origin: Where the dataset files come from.
        """
        super().__init__(size=size, type=DatasetSourceType.FILES, origin=origin, usage=usage)
        self.files = files

    def asdict(self) -> dict:
        return {
            "files": [file.asdict() for file in self.files],
            "size": self.size,
            "type": self.type.value,
            "usage": self.usage.value if self.usage else None,
            "origin": self.origin,
        }


class File:
    """Describe a dataset file."""

    def __init__(
        self,
        name: str,
        size: int,
        fingerprint: str,
        created_date: str | None = None,
        updated_date: str | None = None,
        uri: str | None = None,
    ):
        """Initialize a file.

        Parameters:
            name: The name of the file.
            size: The size of the file.
            fingerprint: The hash of the file.
            created_date: The date of creation of the file.
            updated_date: The date of last update of the file.
            uri: The uri of the file.
        """
        self.name = name
        self.size = size
        self.fingerprint = fingerprint
        self.created_date = created_date
        self.updated_date = updated_date
        self.uri = uri

    def asdict(self) -> dict:
        return {
            "name": self.name,
            "size": self.size,
            "fingerprint": self.fingerprint,
            "createdDate": self.created_date,
            "updatedDate": self.updated_date,
            "uri": self.uri,
        }



from __future__ import annotations

import logging
import os
from contextlib import contextmanager
from io import BytesIO, IOBase
from pathlib import Path
from typing import TYPE_CHECKING, Any

from PIL import Image, UnidentifiedImageError
from rich.console import Console
from rich.table import Table

from vectice.api.json.iteration import IterationStatus, IterationStepArtifact
from vectice.api.json.step import StepType

if TYPE_CHECKING:
    from vectice.models.iteration import Iteration
    from vectice.models.step import Step


@contextmanager
def hide_logs(package: str):
    old_level = logging.getLogger(package).level
    try:
        logging.getLogger(package).setLevel(logging.ERROR)
        yield
    finally:
        logging.getLogger(package).setLevel(old_level)


def _check_read_only(iteration: Iteration):
    """Check if an iteration is completed or canceled.

    Refreshing the iteration is necessary because in a Jupyter notebook
    its status could have changed on the backend.

    Parameters:
        iteration: The iteration to check.

    Raises:
        RuntimeError: When the iteration is read-only (completed or canceled).
    """
    refresh_iteration = iteration._phase.iteration(iteration.index)
    if refresh_iteration._status in {IterationStatus.Completed, IterationStatus.Abandoned}:
        raise RuntimeError(f"The Iteration is {refresh_iteration._status.name} and is read-only!")


def _get_step_type(
    id: int,
    iteration: Iteration,
    name: str,
    index: int,
    slug: str,
    step_type: StepType,
    description: str | None = None,
    completed: bool | None = None,
    artifacts: list[IterationStepArtifact] | None = None,
    text: str | None = None,
) -> Step | Any:
    # TODO: cyclic imports
    from vectice.models.step import Step
    from vectice.models.step_dataset import StepDataset
    from vectice.models.step_image import StepImage
    from vectice.models.step_number import StepNumber
    from vectice.models.step_string import StepString

    step = Step(id, iteration, name, index, slug, description, completed, artifacts, step_type, text)
    if step_type is StepType.StepNumber:
        return StepNumber(step, float(text) if text else None)
    if step_type is StepType.StepString:
        return StepString(step, text)
    if step_type is StepType.StepImage:
        image = _get_image_info(iteration, artifacts) if artifacts else None
        return StepImage(step, image=image)
    if step_type is StepType.StepDataset:
        dataset_version = artifacts[0]  # type: ignore[index]
        return StepDataset(step, dataset_version=dataset_version)
    return step


def _get_image_info(iteration: Iteration, artifacts: list[IterationStepArtifact]):
    artifacts = [image for image in artifacts if image.type.value == "EntityFile"]
    artifact = artifacts[0] if artifacts else None
    attachments = iteration._client.list_phase_attachments(iteration.phase.id, iteration.project.id)
    if not artifact:
        return None
    image = next(
        img for img in attachments.list if img.contentType == "ImageFile" and img.fileId == artifact.entity_file_id
    )
    return image.fileName


def _check_image_path(path: str) -> bool:
    try:
        check_path = Path(path).exists()
    except OSError:
        return False
    _, ext = os.path.splitext(path)
    pillow_extensions = {exten for exten in Image.registered_extensions()}
    if not check_path and ext in pillow_extensions:
        raise ValueError("Check the image path.")
    if not check_path and ext not in pillow_extensions:
        return False
    return True


def _validate_image(path: str) -> BytesIO:
    try:
        byte_io = BytesIO()
        image = Image.open(path)
        image.save(byte_io, image.format)
        byte_io.seek(0)
    except UnidentifiedImageError:
        raise ValueError(
            f"The provided image {path!r} cannot be opened with Pillow, check its format and permissions."
        ) from None
    return byte_io


def _get_image_variables(value: str | IOBase | Image.Image) -> tuple[BytesIO | IOBase, str]:
    if isinstance(value, IOBase):
        image = value
        filename = os.path.basename(value.name)  # type: ignore[attr-defined]
        return image, filename
    if isinstance(value, str):
        image = _validate_image(value)
        filename = os.path.basename(value)
        return image, filename
    if isinstance(value, Image.Image):
        image = value.tobytes()
        filename = os.path.basename(value.filename)
        return image, filename

    raise ValueError("Unsupported image provided.")


def _temp_print(string: str | None = None, table: Table | None = None) -> None:
    console = Console(width=120)
    if string:
        print(string)
        print()
    if table:
        console.print(table)
        print()



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from gql import gql

if TYPE_CHECKING:
    from vectice.api.json import IterationOutput, IterationStepArtifactInput, StepOutput
    from vectice.api.json.step import StepUpdateInput

from gql.transport.exceptions import TransportQueryError

from vectice.api.gql_api import GqlApi, Parser
from vectice.api.json.step import StepType

_logger = logging.getLogger(__name__)

_RETURNS_WITH_STEPS = """
                steps {
                    id
                    index
                    name
                    completed
                    description
                    stepType
                    text
                    slug
                    artifacts {
                                entityFileId
                                modelVersionId
                                datasetVersionId
                                type
                    }
                    __typename
                }
                phase {
                    name
                }
              __typename
            """

_RETURNS = """
                id
                index
                name
                completed
                description
                stepType
                text
                slug
                artifacts {
                                entityFileId
                                modelVersionId
                                datasetVersionId
                                type
                    }
                __typename
            """


class StepApi(GqlApi):
    def list_steps(self, phase_id: int, phase_name: str | None) -> list[StepOutput]:
        gql_query = "getActiveIterationOrCreateOne"
        variable_types = "$phaseId:Float!"
        variables = {"phaseId": phase_id}
        kw = "phaseId:$phaseId"
        query = GqlApi.build_query(
            gql_query=gql_query,
            variable_types=variable_types,
            returns=_RETURNS_WITH_STEPS,
            keyword_arguments=kw,
            query=True,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            iteration_output: IterationOutput = Parser().parse_item(response[gql_query])
            step_output: list[StepOutput] = iteration_output.steps
            return step_output
        except TransportQueryError as e:
            phase_ref: int | str = phase_name if phase_name else phase_id
            self._error_handler.handle_post_gql_error(e, "steps", phase_ref)

    def list_steps_for_iteration(
        self, phase_id: int, iteration_index: int, phase_name: str | None = None
    ) -> list[StepOutput]:
        gql_query = "getIterationByIndex"
        variable_types = "$index:Float!,$phaseId:Float!"
        variables = {"index": iteration_index, "phaseId": phase_id}
        kw = "index:$index,phaseId:$phaseId"
        query = GqlApi.build_query(
            gql_query=gql_query,
            variable_types=variable_types,
            returns=_RETURNS_WITH_STEPS,
            keyword_arguments=kw,
            query=True,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            iteration_output: IterationOutput = Parser().parse_item(response[gql_query])
            step_output: list[StepOutput] = iteration_output.steps
            return step_output
        except TransportQueryError as e:
            phase_ref: int | str = phase_name if phase_name else phase_id
            self._error_handler.handle_post_gql_error(e, "steps", phase_ref)

    def close_step(self, step_id: int, step_update: StepUpdateInput) -> StepOutput:
        gql_query = "completeIterationStep"
        variable_types = "$id:Float!,$data:IterationStepUpdateInput"
        variables = {"id": step_id, "data": step_update}
        kw = "id:$id, data:$data"
        query = GqlApi.build_query(
            gql_query=gql_query,
            variable_types=variable_types,
            returns=_RETURNS,
            keyword_arguments=kw,
            query=False,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            step_output: StepOutput = Parser().parse_item(response[gql_query])
            return step_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "step", step_id)

    def add_iteration_step_artifact(self, data: IterationStepArtifactInput, step_id: int) -> StepOutput:
        gql_query = "addIterationStepArtifact"
        variable_types = "$id:Float!,$data:IterationStepArtifactInput!"
        variables = {"id": step_id, "data": data}
        kw = "id:$id, data:$data"
        query = GqlApi.build_query(
            gql_query=gql_query,
            variable_types=variable_types,
            returns=_RETURNS,
            keyword_arguments=kw,
            query=False,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            step_output: StepOutput = Parser().parse_item(response[gql_query])
            return step_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "step", step_id)

    def update_iteration_step_artifact(
        self,
        step_id: int,
        step_type: StepType,
        text: str | None = None,
        artifacts: list[IterationStepArtifactInput] | None = None,
    ) -> StepOutput:
        gql_query = "updateIterationStepContent"
        variable_types = "$id:Float!,$data:IterationStepUpdateInput!"
        variables: dict = {
            "id": step_id,
            "data": {
                "stepType": step_type.value,
                "text": str(text) if text else None,
            },
        }
        if artifacts:
            variables["data"]["artifacts"] = artifacts
        kw = "id:$id, data:$data"
        query = GqlApi.build_query(
            gql_query=gql_query,
            variable_types=variable_types,
            returns=_RETURNS,
            keyword_arguments=kw,
            query=False,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            step_output: StepOutput = Parser().parse_item(response[gql_query])
            return step_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "step", step_id)

    def get_step(self, step: str, iteration_id: int) -> StepOutput:
        if isinstance(step, str) and iteration_id:
            gql_query = "getStepByName"
            variable_types = "$name:String!,$parentId:Float!"
            variables = {"name": step, "parentId": iteration_id}
            kw = "name:$name,parentId:$parentId"
        else:
            raise ValueError("string and parent id required.")
        query = GqlApi.build_query(
            gql_query=gql_query, variable_types=variable_types, returns=_RETURNS, keyword_arguments=kw, query=True
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            step_output: StepOutput = Parser().parse_item(response[gql_query])
            return step_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "step", step)



from __future__ import annotations

from enum import Enum


class PhaseStatus(Enum):
    """Enumeration of the different phase statuses."""

    NotStarted = "NotStarted"
    InProgress = "Draft"
    Completed = "Completed"
    InReview = "InReview"


class PhaseInput(dict):
    pass


class PhaseOutput(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def id(self) -> int:
        return int(self["id"])

    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def index(self) -> int:
        return int(self["index"])

    @property
    def status(self) -> PhaseStatus:
        return PhaseStatus(self["status"])



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from gql import gql
from gql.transport.exceptions import TransportQueryError

from vectice.api.gql_api import GqlApi, Parser

if TYPE_CHECKING:
    from vectice.api.json.phase import PhaseOutput
    from vectice.api.json.step import StepOutput


_RETURNS = """id
              name
              status
              index
              owner {
                    name
              }
              __typename
            """

_STEP_DEFINITION_RETURNS = """id
              name
              index
              slug
              description
              __typename
            """


_logger = logging.getLogger(__name__)


class PhaseApi(GqlApi):
    def list_phases(self, parent_id: int, search_alias: str | None = None) -> list[PhaseOutput]:
        alias_filter = {
            "parentId": parent_id,
            "searchFilter": {"search": search_alias if search_alias is not None else "", "fields": "name"},
        }

        variable_types = "$filters:PhaseFiltersInput!"
        kw = "filters:$filters"
        variables = {"filters": alias_filter}
        returns = f"""items{{
                    {_RETURNS}
        }}"""
        query = GqlApi.build_query(
            gql_query="getPhaseList", variable_types=variable_types, returns=returns, keyword_arguments=kw, query=True
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            phase_output: list[PhaseOutput] = Parser().parse(response["getPhaseList"]["items"])  # type: ignore[assignment]
            return phase_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "phase", "list")

    def get_phase(self, phase: str | int, parent_id: int | None = None) -> PhaseOutput:
        if isinstance(phase, int):
            gql_query = "getPhaseById"
            variable_types = "$id:Float!"
            variables = {"id": phase}
            kw = "id:$id"
        elif isinstance(phase, str) and parent_id:
            gql_query = "getPhaseByName"
            variable_types = "$name:String!,$parentId:Float!"
            variables = {"name": phase, "parentId": parent_id}  # type: ignore[dict-item]
            kw = "name:$name,parentId:$parentId"
        else:
            raise ValueError("string and parent id required.")
        query = GqlApi.build_query(
            gql_query=gql_query, variable_types=variable_types, returns=_RETURNS, keyword_arguments=kw, query=True
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            phase_output: PhaseOutput = Parser().parse_item(response[gql_query])
            return phase_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "phase", phase)

    def list_step_definitions(self, parent_id: int) -> list[StepOutput]:
        alias_filter = {"parentId": parent_id, "search": ""}
        returns = f"""items{{
                                    {_STEP_DEFINITION_RETURNS}
                        }}"""
        variable_types = "$filters:BaseDocumentationListFiltersInput!"
        kw = "filters:$filters"
        variables = {"filters": alias_filter}
        query = GqlApi.build_query(
            gql_query="getStepDefinitionList",
            variable_types=variable_types,
            returns=returns,
            keyword_arguments=kw,
            query=True,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            iterations_output: list[StepOutput] = Parser().parse_list(response["getStepDefinitionList"]["items"])
            return iterations_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "phase", "list")



from __future__ import annotations

from typing import Any

from vectice.models.resource.metadata.base import (
    DatasetSourceOrigin,
    DatasetSourceType,
    DatasetSourceUsage,
    Metadata,
)
from vectice.models.resource.metadata.files_metadata import (
    File,
    FilesMetadata,
)
from vectice.utils.deprecation import deprecate


def __getattr__(name: str) -> Any:
    attrs = {
        "DatasetSourceOrigin": DatasetSourceOrigin,
        "DatasetSourceType": DatasetSourceType,
        "DatasetSourceUsage": DatasetSourceUsage,
        "Metadata": Metadata,
        "FilesMetadata": FilesMetadata,
        "File": File,
    }

    if name in attrs:
        return attrs[name]

    raise AttributeError(f"'{__name__}' module has no attribute '{name}'")


@deprecate(
    warn_at="23.1",
    fail_at="23.3",
    remove_at="23.4",
    reason="The `vectice.models.datasource.datawrapper.metadata.files_metadata` module "
    "is deprecated in favor the `vectice.models.resource.metadata.files_metadata` module. "
    "Importing from the deprecated module will fail in v{fail_at}, "
    "and the module will be removed in v{remove_at}.",
)
def _warn() -> None:
    pass


_warn()



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from gql import gql
from gql.transport.exceptions import TransportQueryError

from vectice.api.gql_api import GqlApi, Parser

if TYPE_CHECKING:
    from vectice.api.json.iteration import IterationInput, IterationOutput

_logger = logging.getLogger(__name__)

_RETURNS = """id
            index
            status
            phase {
                    id
                    name
                    status
                    __typename
              }
            steps {
                    slug
                    id
                    name
                    stepType
                    artifacts {
                    type
                    datasetVersionId
                    }
            }
            __typename
            """

_RETURNS_LAST_ASSETS = """id
            index
            status
            phase {
                    id
                    name
                    status
                    __typename
                    parent {
                            id
                            name
                            __typename
                            workspace{
                                    id
                                    name
                                    __typename
                            }
                    }
              }
              steps{
                    slug
                    id
                    name
                    index
                    completed
                    updatedDate
                    __typename
              }
            __typename
            """


class IterationApi(GqlApi):
    def list_iterations(self, parent_id: int) -> list[IterationOutput]:
        alias_filter = {"phaseId": parent_id, "search": ""}
        returns = f"""items{{
                                    {_RETURNS}
                        }}"""
        variable_types = "$filters:IterationFiltersInput!"
        kw = "filters:$filters"
        variables = {"filters": alias_filter}
        query = GqlApi.build_query(
            gql_query="getIterationList",
            variable_types=variable_types,
            returns=returns,
            keyword_arguments=kw,
            query=True,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            iterations_output: list[IterationOutput] = Parser().parse_list(response["getIterationList"]["items"])
            return iterations_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "iteration", "list")

    def create_iteration(self, phase_id: int) -> IterationOutput:
        gql_query = "createIteration"
        variable_types = "$phaseId:Float!"
        variables = {"phaseId": phase_id}
        kw = "phaseId:$phaseId"
        query = GqlApi.build_query(
            gql_query=gql_query, variable_types=variable_types, returns=_RETURNS, keyword_arguments=kw, query=False
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            iteration_output: IterationOutput = Parser().parse_item(response[gql_query])
            return iteration_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "iteration", phase_id)

    def get_iteration(self, iteration_id: int) -> IterationOutput:
        gql_query = "getIterationById"
        variable_types = "$id:Float!"
        variables = {"id": iteration_id}
        kw = "id:$id"
        query = GqlApi.build_query(
            gql_query=gql_query, variable_types=variable_types, returns=_RETURNS, keyword_arguments=kw, query=True
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            iteration_output: IterationOutput = Parser().parse_item(response[gql_query])
            return iteration_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "iteration", iteration_id)

    def get_iteration_last_assets(self, iteration_id: int) -> IterationOutput:
        gql_query = "getIterationById"
        variable_types = "$id:Float!"
        variables = {"id": iteration_id}
        kw = "id:$id"
        query = GqlApi.build_query(
            gql_query=gql_query,
            variable_types=variable_types,
            returns=_RETURNS_LAST_ASSETS,
            keyword_arguments=kw,
            query=True,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            iteration_output: IterationOutput = Parser().parse_item(response[gql_query])
            return iteration_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "iteration", iteration_id)

    def get_iteration_by_index(self, phase_id: int, index: int) -> IterationOutput:
        gql_query = "getIterationByIndex"
        variable_types = "$index:Float!,$phaseId:Float!"
        variables = {"index": index, "phaseId": phase_id}
        kw = "index:$index,phaseId:$phaseId"
        query = GqlApi.build_query(
            gql_query=gql_query, variable_types=variable_types, returns=_RETURNS, keyword_arguments=kw, query=True
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            iteration_output: IterationOutput = Parser().parse_item(response[gql_query])
            return iteration_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "iteration_index", index)

    def update_iteration(self, iteration: IterationInput, iteration_id: int) -> IterationOutput:
        variable_types = "$id:Float!,$data:IterationUpdateInput!"
        kw = "id:$id,data:$data"
        variables = {"id": iteration_id, "data": iteration}
        query = GqlApi.build_query(
            gql_query="updateIteration",
            variable_types=variable_types,
            returns=_RETURNS,
            keyword_arguments=kw,
            query=False,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            iteration_output: IterationOutput = Parser().parse_item(response["updateIteration"])
            return iteration_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "iteration", "put")

    def delete_iteration(self, iteration_id: int) -> None:
        variable_types = "$id:Float!"
        kw = "id:$id"
        variables = {"id": iteration_id}
        query = GqlApi.build_query(
            gql_query="removeIteration",
            variable_types=variable_types,
            keyword_arguments=kw,
            query=False,
        )
        query_built = gql(query)
        try:
            self.execute(query_built, variables)
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "iteration", iteration_id)



from __future__ import annotations

from vectice.models.datasource.datawrapper.data_wrapper import DataWrapper
from vectice.models.datasource.datawrapper.file_data_wrapper import File, FileDataWrapper, FilesMetadata
from vectice.models.datasource.datawrapper.gcs_data_wrapper import GcsDataWrapper, NoSuchGcsResourceError
from vectice.models.datasource.datawrapper.s3_data_wrapper import NoSuchS3ResourceError, S3DataWrapper

__all__ = [
    "DataWrapper",
    "FilesMetadata",
    "FileDataWrapper",
    "File",
    "GcsDataWrapper",
    "NoSuchGcsResourceError",
    "S3DataWrapper",
    "NoSuchS3ResourceError",
]



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from gql import gql
from gql.transport.exceptions import TransportQueryError

from vectice.api.gql_api import GqlApi, Parser

if TYPE_CHECKING:
    from vectice.api.json.dataset_register import DatasetRegisterInput, DatasetRegisterOutput


_logger = logging.getLogger(__name__)

# TODO JobRun for lineages
_RETURNS = """
            datasetVersion{
                          id
                          name
            }
            useExistingDataset
            useExistingVersion
            __typename
            """


class GqlDatasetApi(GqlApi):
    def register_dataset(
        self,
        data: DatasetRegisterInput,
        project_id: int | None = None,
        phase_id: int | None = None,
        iteration_id: int | None = None,
    ) -> DatasetRegisterOutput:
        if phase_id and project_id and not iteration_id:
            variable_types = "$projectId:Float!,$phaseId:Float!,$data:DatasetRegisterInput!"
            kw = "projectId:$projectId,phaseId:$phaseId,data:$data"
            variables = {"projectId": project_id, "phaseId": phase_id, "data": data}
        elif project_id and phase_id and iteration_id:
            variable_types = "$projectId:Float!,$phaseId:Float!,$iterationId:Float!,$data:DatasetRegisterInput!"
            kw = "projectId:$projectId,phaseId:$phaseId,iterationId:$iterationId,data:$data"
            variables = {"projectId": project_id, "phaseId": phase_id, "iterationId": iteration_id, "data": data}
        elif project_id:
            variable_types = "$projectId:Float!,$data:DatasetRegisterInput!"
            kw = "projectId:$projectId,data:$data"
            variables = {"projectId": project_id, "data": data}
        else:
            raise RuntimeError("The provided parent child ids do not match!")

        query = GqlApi.build_query(
            gql_query="registerDataset",
            variable_types=variable_types,
            returns=_RETURNS,
            keyword_arguments=kw,
            query=False,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            dataset_output: DatasetRegisterOutput = Parser().parse_item(response["registerDataset"])
            return dataset_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "dataset", "register dataset")



from __future__ import annotations

import os
from datetime import datetime

from dotenv import dotenv_values, find_dotenv


def read_nodejs_date(date_as_string: str | None) -> datetime | None:
    if date_as_string is None:
        return None
    return datetime.strptime(date_as_string, "%Y-%m-%dT%H:%M:%S.%f%z")


def calculate_duration(end_date: datetime, start_date_nodejs_format: str) -> int:
    # format sample : 2021-06-20T11:04:16.249Z
    start_date = read_nodejs_date(start_date_nodejs_format)
    if start_date is None:
        raise RuntimeError("Invalid date format for value: " + start_date_nodejs_format)
    duration = end_date - start_date
    return int(duration.total_seconds())


def read_env(*args: str) -> list[str | None]:
    """Read user configuration.

    Configuration is read from different locations, in this order:

    - `.vectice` file
    - `.env` file
    - environment variables

    As usual, system variables always take precedence over dotenv files.
    And as expected, dotenv files take precedence over dotvectice files.

    Parameters:
        *args: The keys of the configuration items to return.

    Raises:
        ValueError: When no arguments are given.

    Returns:
        A list of configuration items.
    """
    if len(args) == 0:
        raise ValueError("At least one argument must be provided to read.")
    config: dict[str, str | None] = {
        **dotenv_values(find_dotenv(".vectice")),
        **dotenv_values(find_dotenv(".env")),
        **os.environ,
    }
    return [config.get(key) for key in args]



from __future__ import annotations

from datetime import datetime

from vectice.models.resource.base import Resource
from vectice.models.resource.metadata.base import DatasetSourceUsage, DatasetType


class Dataset:
    def __init__(
        self,
        type: DatasetType,
        name: str | None = None,
        resource: Resource | None = None,
        training_resource: Resource | None = None,
        testing_resource: Resource | None = None,
        validation_resource: Resource | None = None,
        derived_from: list[int | Dataset] | None = None,
        # attachments: str | list[str] | None = None,
        # properties: dict[str, str | int] | list[Property] | Property | None = None,
    ):
        """Initialize a dataset.

        Users should not instantiate a dataset directly but rather use the provided static methods
        [`origin()`][vectice.models.dataset.Dataset.origin],
        [`clean()`][vectice.models.dataset.Dataset.clean], and
        [`modeling()`][vectice.models.dataset.Dataset.modeling].

        Parameters:
            type: The type of dataset.
            name: The name of the dataset.
            resource: A single resource (for origin and clean datasets).
            training_resource: The resource for the training set (for modeling datasets).
            testing_resource: The resource for the testing set (for modeling datasets).
            validation_resource: The resource for the validation set (optional, for modeling datasets).
            derived_from: A list of datasets (or ids) from which this dataset is derived.
        """
        derived_from_ids = []
        for df in derived_from or []:
            if isinstance(df, Dataset):
                if df.latest_version_id is None:
                    raise ValueError(
                        f"Dataset '{df.name}' does not have a version id. "
                        "Was it registered in Vectice (assigned to a step)?"
                    )
                derived_from_ids.append(df.latest_version_id)
            else:
                derived_from_ids.append(df)
        self._type = type
        self._name = name or f"dataset {datetime.time}"
        self._resource = resource
        self._training_resource = training_resource
        self._testing_resource = testing_resource
        self._validation_resource = validation_resource
        self._derived_from = derived_from_ids
        self._latest_version_id: int | None = None
        # self._properties = self._format_properties(properties) if properties else None
        # self._attachments = self._format_attachments(attachments) if attachments else None

        if self._type is DatasetType.MODELING:
            if self._training_resource is None or self._testing_resource is None:
                raise ValueError("You cannot create a modeling dataset without both training and testing sets")
            self._training_resource.usage = DatasetSourceUsage.TRAINING
            self._testing_resource.usage = DatasetSourceUsage.TESTING
            if self._validation_resource:
                self._validation_resource.usage = DatasetSourceUsage.VALIDATION

    @staticmethod
    def origin(
        resource: Resource,
        name: str | None = None,
        # properties: dict[str, str | int] | list[Property] | Property | None = None,
        # attachments: str | list[str] | None = None,
    ) -> Dataset:
        """Create an origin dataset.

        Examples:
            ```python
            from vectice import Dataset, FileResource

            dataset = Dataset.origin(
                name="my origin dataset",
                resource=FileResource(path="origin_dataset.csv"),
            )
            ```

        Parameters:
            resource: The resource for the origin dataset.
            name: The name of the dataset.
        """
        return Dataset(
            type=DatasetType.ORIGIN,
            name=name,
            resource=resource,
            # properties=properties,
            # attachments=attachments,
        )

    @staticmethod
    def clean(
        resource: Resource,
        name: str | None = None,
        derived_from: list[int | Dataset] | None = None,
        # properties: dict[str, str | int] | list[Property] | Property | None = None,
        # attachments: str | list[str] | None = None,
    ) -> Dataset:
        """Create a clean dataset.

        Examples:
            ```python
            from vectice import Dataset, FileResource

            dataset = Dataset.clean(
                name="my clean dataset",
                resource=FileResource(path="clean_dataset.csv"),
            )
            ```

        Parameters:
            resource: The resource for the clean dataset.
            name: The name of the dataset.
            derived_from: A list of datasets (or ids) from which this dataset is derived.
        """
        return Dataset(
            type=DatasetType.CLEAN,
            name=name,
            resource=resource,
            derived_from=derived_from,
            # properties=properties,
            # attachments=attachments,
        )

    @staticmethod
    def modeling(
        training_resource: Resource,
        testing_resource: Resource,
        validation_resource: Resource | None = None,
        name: str | None = None,
        # properties: dict[str, str | int] | list[Property] | Property | None = None,
        # attachments: str | list[str] | None = None,
    ) -> Dataset:
        """Create a modeling dataset.

        Examples:
            ```python
            from vectice import Dataset, FileResource

            dataset = Dataset.modeling(
                name="my modeling dataset",
                training_resource=FileResource(path="training_dataset.csv"),
                testing_resource=FileResource(path="testing_dataset.csv"),
                validation_resource=FileResource(path="validation_dataset.csv"),
            )
            ```

        Parameters:
            training_resource: The resource for the training set (for modeling datasets).
            testing_resource: The resource for the testing set (for modeling datasets).
            validation_resource: The resource for the validation set (optional, for modeling datasets).
            name: The name of the dataset.
        """
        return Dataset(
            type=DatasetType.MODELING,
            name=name,
            training_resource=training_resource,
            testing_resource=testing_resource,
            validation_resource=validation_resource,
            # properties=properties,
            # attachments=attachments,
        )

    def __repr__(self):
        return f"Dataset(name={self.name!r}, type={self.type!r})"

    @property
    def type(self) -> DatasetType:
        """The dataset's type.

        Returns:
            The dataset's type.
        """
        return self._type

    @property
    def resource(self) -> Resource | tuple[Resource, Resource, Resource | None]:
        """The dataset's resource.

        Returns:
            The dataset's resource.
        """
        if self._type is DatasetType.MODELING:
            return self._training_resource, self._testing_resource, self._validation_resource  # type: ignore[return-value]
        return self._resource  # type: ignore[return-value]

    @property
    def name(self) -> str:
        """The dataset's name.

        Returns:
            The dataset's name.
        """
        return self._name

    @name.setter
    def name(self, name):
        """Set the dataset's name.

        Parameters:
            name: The name of the dataset.
        """
        self._name = name

    @property
    def derived_from(self) -> list[int]:
        """The datasets from which this dataset is derived.

        Returns:
            The datasets from which this dataset is derived.
        """
        return self._derived_from

    @property
    def latest_version_id(self) -> int | None:
        """The id of the latest version of this dataset.

        Returns:
            The id of the latest version of this dataset.
        """
        return self._latest_version_id

    @latest_version_id.setter
    def latest_version_id(self, value: int) -> None:
        """Set the id of the latest version of this dataset.

        Parameters:
            value: The id of the latest version of this dataset.
        """
        self._latest_version_id = value

    # @property
    # def properties(self) -> list[Property] | None:
    #     """The dataset's properties.

    #     Returns:
    #         The dataset's properties.
    #     """
    #     return self._properties

    # @properties.setter
    # def properties(self, properties: dict[str, str | int] | list[Property] | Property | None):
    #     """Set the dataset's properties.

    #     Parameters:
    #         properties: The properties of the dataset.
    #     """
    #     self._properties = self._format_properties(properties)

    # @property
    # def attachments(self) -> list[str] | None:
    #     """The attachments associated with the dataset.

    #     Returns:
    #         The attachments associated with the dataset.
    #     """
    #     return self._attachments

    # @attachments.setter
    # def attachments(self, attachments: list[str] | str):
    #     """Attach a file or files to the dataset.

    #     Parameters:
    #         attachments: The filename or filenames of the file or set of files to attach to the dataset.
    #     """
    #     self._attachments = self._format_attachments(attachments)

    # @staticmethod
    # def _check_key_duplicates(key_list: list[str]):
    #     if len(key_list) != len(set(key_list)):
    #         raise ValueError("Duplicate keys are not allowed.")

    # @staticmethod
    # def _format_attachments(attachments: str | list[str]) -> list[str]:
    #     return [attachment for attachment in set(attachments)] if isinstance(attachments, list) else [attachments]

    # def _format_properties(self, properties: dict[str, str | int] | list[Property] | Property | None) -> list[Property]:
    #     if properties is None:
    #         return []
    #     if isinstance(properties, Property):
    #         return [properties]
    #     if isinstance(properties, list):
    #         properties = self._remove_incorrect_properties(properties)
    #         key_list = [prop.key for prop in properties]
    #         self._check_key_duplicates(key_list)
    #         return properties
    #     if isinstance(properties, dict):
    #         return [Property(key, str(value)) for (key, value) in properties.items()]
    #     raise ValueError("Please check property type.")

    # @staticmethod
    # def _remove_incorrect_properties(properties: list[Property]) -> list[Property]:
    #     for prop in properties:
    #         if not isinstance(prop, Property):
    #             logging.warning(f"Incorrect property '{prop}'. Please check property type.")
    #             properties.remove(prop)
    #     return properties



from __future__ import annotations

import logging
from datetime import datetime
from typing import TYPE_CHECKING, BinaryIO

from gql import Client as GQLClient
from gql.transport.requests import RequestsHTTPTransport

from vectice.__version__ import __version__
from vectice.api._auth import Auth
from vectice.api.attachment import AttachmentApi
from vectice.api.compatibility import CompatibilityApi
from vectice.api.gql_code import GqlCodeApi
from vectice.api.gql_code_version import GqlCodeVersionApi
from vectice.api.gql_dataset import GqlDatasetApi
from vectice.api.gql_model import GqlModelApi
from vectice.api.gql_user_workspace_api import UserAndDefaultWorkspaceApi
from vectice.api.http_error_handlers import MissingReferenceError, StepIdError, StepNameError
from vectice.api.iteration import IterationApi
from vectice.api.json import (
    ArtifactName,
    CodeInput,
    CodeVersionCreateBody,
    ModelRegisterInput,
    ModelRegisterOutput,
    ModelType,
    ModelVersionOutput,
    ModelVersionStatus,
    Page,
    PagedResponse,
    ProjectInput,
    PropertyInput,
    StepOutput,
)
from vectice.api.json.dataset_register import DatasetRegisterInput, DatasetRegisterOutput
from vectice.api.json.metric import MetricInput
from vectice.api.json.step import StepUpdateInput
from vectice.api.last_assets import LastAssetApi
from vectice.api.phase import PhaseApi
from vectice.api.project import ProjectApi
from vectice.api.step import StepApi
from vectice.api.version import VersionApi
from vectice.api.workspace import WorkspaceApi
from vectice.models.dataset import Dataset
from vectice.utils.deprecation import deprecate

if TYPE_CHECKING:
    from io import BytesIO, IOBase

    from requests import Response

    from vectice.api.json import AttachmentOutput, ProjectOutput, WorkspaceOutput
    from vectice.api.json.compatibility import CompatibilityOutput
    from vectice.api.json.iteration import IterationInput, IterationOutput, IterationStepArtifactInput
    from vectice.api.json.phase import PhaseOutput
    from vectice.api.json.step import StepType
    from vectice.api.json.workspace import WorkspaceInput
    from vectice.models.model import Model


_logger = logging.getLogger(__name__)


def _auth_success_message(workspace: WorkspaceOutput, project: ProjectOutput | None) -> str:
    if workspace and project:
        return f"Successfully authenticated. You'll be working on Project: '{project.name}', part of Workspace: '{workspace.name}'"
    return f"Successfully authenticated. Your current Workspace: '{workspace.name}'"


class Client:
    """Low level Vectice API client."""

    def __init__(
        self,
        workspace: str | int | None = None,
        project: str | int | None = None,
        token: str | None = None,
        api_endpoint: str | None = None,
        auto_connect=True,
        allow_self_certificate=True,
    ):
        self.auth = Auth(
            api_endpoint=api_endpoint,
            api_token=token,
            auto_connect=auto_connect,
            allow_self_certificate=allow_self_certificate,
        )
        transport = RequestsHTTPTransport(url=self.auth.api_base_url + "/graphql", verify=self.auth.verify_certificate)
        logging.getLogger("gql.transport.requests").setLevel("WARNING")
        self._gql_client = GQLClient(transport=transport)
        self._logger = logging.getLogger(self.__class__.__name__)
        self._workspace = None
        self._project = None

        if auto_connect:
            if workspace and project:
                self._project = self.get_project(project, workspace)
                self._workspace = self._project.workspace
                if (
                    isinstance(workspace, str)
                    and workspace != self._workspace.name
                    or isinstance(workspace, int)
                    and workspace != self._workspace.id
                ):
                    raise ValueError(
                        f"Inconsistency in configuration: Project {project} does not belong to Workspace {workspace}"
                    )
            elif workspace:
                self._workspace = self.get_workspace(workspace)
            elif project:
                self._project = self.get_project(project)
                self._workspace = self._project.workspace

            if self._workspace:
                _logger.debug(_auth_success_message(workspace=self._workspace, project=self._project))

    @property
    def workspace(self) -> WorkspaceOutput | None:
        """The workspace object.

        Returns:
            The workspace object.
        """
        return self._workspace

    @property
    def project(self) -> ProjectOutput | None:
        """The project object.

        Returns:
            The project object.
        """
        return self._project

    @property
    def version_api(self) -> str:
        return __version__

    @property
    def version_backend(self) -> str:
        versions = VersionApi(self._gql_client, self.auth).get_public_config().versions
        for version in versions:
            if version.artifact_name == ArtifactName.BACKEND:
                return version.version
        raise ValueError("No version found for backend.")

    def check_compatibility(self) -> CompatibilityOutput:
        return CompatibilityApi(self.auth).check_version()

    def create_project(self, data: ProjectInput, workspace: str | int) -> ProjectOutput:
        """Create a project.

        Parameters:
            data: The ProjectInput JSON structure.
            workspace: The workspace name or id.

        Returns:
            The project JSON structure.
        """
        result = ProjectApi(self.auth).create_project(data, workspace)
        _logger.info(f"Project with id: {result.id} successfully created.")
        return result

    def delete_project(self, project: str | int, workspace: str | int | None = None):
        """Delete a project.

        Parameters:
            project: The project name or id.
            workspace: The workspace name or id.
        """
        ProjectApi(self.auth).delete_project(project, workspace)

    def update_project(self, data: ProjectInput, project: str | int, workspace: str | int) -> ProjectOutput:
        """Update a project.

        Parameters:
            data: The ProjectInput JSON structure.
            project: The project name or id.
            workspace: The workspace name or id.

        Returns:
            The project JSON structure.
        """
        return ProjectApi(self.auth).update_project(data, project, workspace)

    def list_projects(
        self,
        workspace: str | int,
        search: str | None = None,
        page_index: int | None = Page.index,
        page_size: int | None = Page.size,
    ) -> PagedResponse[ProjectOutput]:
        """List the projects in a workspace.

        Parameters:
            workspace: The workspace name or id.
            search: A text to search for.
            page_index: The index of the page.
            page_size: The size of the page.

        Returns:
            The workspace's projects.
        """
        return ProjectApi(self.auth).list_projects(workspace, search, page_index, page_size)

    def get_project(self, project: str | int, workspace: str | int | None = None) -> ProjectOutput:
        """Get a project.

        Parameters:
            project: The project name or id.
            workspace: The workspace name or id.

        Returns:
            The project JSON structure.
        """
        return ProjectApi(self.auth).get_project(project, workspace)

    def get_workspace(self, workspace: str | int) -> WorkspaceOutput:
        """Get a workspace.

        Parameters:
            workspace: The workspace name or id.

        Returns:
            The workspace JSON structure.
        """
        return WorkspaceApi(self.auth).get_workspace(workspace)

    def create_workspace(self, data: WorkspaceInput) -> WorkspaceOutput:
        """Create a workspace.

        Parameters:
            data: The WorkspaceInput JSON structure.

        Returns:
            The workspace JSON structure.
        """
        result = WorkspaceApi(self.auth).create_workspace(data)
        return result

    def update_workspace(self, data: WorkspaceInput, workspace: str | int) -> WorkspaceOutput:
        """Update a workspace.

        Parameters:
            data: The WorkspaceInput JSON structure.
            workspace: The workspace name or id.

        Returns:
            The workspace JSON structure.
        """
        return WorkspaceApi(self.auth).update_workspace(data, workspace)

    def list_workspaces(
        self, search: str | None = None, page_index: int = 1, page_size: int = 20
    ) -> PagedResponse[WorkspaceOutput]:
        """List the workspaces.

        Parameters:
            search: A text to search for.
            page_index: The index of the page.
            page_size: The size of the page.

        Returns:
            The workspaces.
        """
        return WorkspaceApi(self.auth).list_workspaces(search, page_index, page_size)

    def create_code_attachments(self, files: list[tuple[str, tuple[str, str]]], code_version_id: int, project_id: int):
        """Create an attachment.

        Parameters:
            files: The paths to the files to attach.
            code_version_id: The code version id to attach files to.
            project_id: The project id associated to the code version id.

        Returns:
            The JSON structure.
        """
        return AttachmentApi(self.auth).create_attachments(files, project_id, code_version_id=code_version_id)

    def create_model_attachments(
        self, files: list[tuple[str, tuple[str, BinaryIO]]], model_version: ModelVersionOutput
    ):
        """Create an attachment.

        Parameters:
            files: The paths to the files to attach.
            model_version: The model version to attach files to.

        Returns:
            The JSON structure.
        """
        return AttachmentApi(self.auth).post_attachment(files, model_version)

    def create_phase_attachments(
        self, files: list[tuple[str, tuple[str, BytesIO | IOBase]]], phase_id: int, project_id: int
    ) -> list[dict]:
        """Create an attachment.

        Parameters:
            files: The paths to the files to attach.
            phase_id: The phase id to attach files to.
            project_id: The project id to attach files to.

        Returns:
            The JSON structure.
        """
        return AttachmentApi(self.auth).create_attachments(files, project_id, phase_id=phase_id)

    def create_model_predictor(self, model_type: str, model_content: BytesIO, model_version: ModelVersionOutput):
        """Create a predictor.

        Parameters:
            model_type: The type of model to attach.
            model_content: The binary content of the model.
            model_version: The model version to attach files to.

        Returns:
            The JSON structure.
        """
        return AttachmentApi(self.auth).post_model_predictor(model_type, model_content, model_version)

    def list_attachments(self, model_version: ModelVersionOutput) -> PagedResponse[AttachmentOutput]:
        """List the attachments of an artifact.

        Parameters:
            model_version: The model version to list attachments from.

        Returns:
            The attachments of an artifact.
        """
        return AttachmentApi(self.auth).list_attachments(model_version)

    def list_code_attachments(self, code_version_id: int, project_id: int) -> PagedResponse[AttachmentOutput]:
        """List the attachments of a code version.

        Parameters:
            code_version_id: The id of the code version to list attachments from.
            project_id: The id of the project the code version belongs to.

        Returns:
            A list of attachments that belong to the code version.
        """
        return AttachmentApi(self.auth).list_object_attachments(project_id, code_version_id=code_version_id)

    def list_phase_attachments(self, phase_id: int, project_id: int) -> PagedResponse[AttachmentOutput]:
        """List the attachments of a phase.

        Parameters:
            phase_id: The id of the phase the attachments belongs to.
            project_id: The id of the project the attachments belongs to.

        Returns:
            An attachment.
        """
        return AttachmentApi(self.auth).list_object_attachments(project_id, phase_id=phase_id)

    def get_code_version_attachment(self, code_version_id: int, project_id: int, file_id: int) -> Response:
        """Get the attachment of a code version.

        Parameters:
            code_version_id: The code version id to list attachments from.
            project_id: The project id the code version belongs to.
            file_id: The file id attached to the code version.

        Returns:
            The file attached to the code version.
        """
        return AttachmentApi(self.auth).get_code_version_attachment(code_version_id, project_id, file_id)

    def list_phases(
        self,
        search: str | None = None,
        project: str | int | None = None,
        workspace: str | int | None = None,
    ) -> list[PhaseOutput]:
        project, workspace = self.get_project_and_workspace_references_or_raise_error(project, workspace)
        project_object = self.get_project(project, workspace)
        return PhaseApi(self._gql_client, self.auth).list_phases(project_object.id, search)

    def get_phase(self, phase: str | int, project_id: int | None = None) -> PhaseOutput:
        if project_id is None:
            raise MissingReferenceError("project")
        return PhaseApi(self._gql_client, self.auth).get_phase(phase, project_id)

    def get_step(self, step_reference: str | int, phase_id: int, iteration_id: int) -> StepOutput:
        if phase_id is None:
            raise MissingReferenceError("iteration")
        steps = self.list_steps(phase_id, iteration_id)
        if isinstance(step_reference, int):
            for step in steps:
                if step.id == step_reference:
                    return step
            raise StepIdError(step_reference)
        elif isinstance(step_reference, str):
            for step in steps:
                if step.name == step_reference:
                    return step
            raise StepNameError(step_reference)
        raise ValueError(f"No such step reference '{step_reference}' exists in the phase '{phase_id}'")

    def get_step_by_name(self, step_reference: str, iteration_id: int) -> StepOutput:
        return StepApi(self._gql_client, self.auth).get_step(step_reference, iteration_id)

    def list_steps(
        self,
        phase_id: int,
        iteration_index: int | None = None,
        phase_name: str | None = None,
    ) -> list[StepOutput]:
        if iteration_index:
            return StepApi(self._gql_client, self.auth).list_steps_for_iteration(phase_id, iteration_index, phase_name)
        return StepApi(self._gql_client, self.auth).list_steps(phase_id, phase_name)

    def close_step(self, step_id: int, message: str | None = None) -> StepOutput:
        step_update = StepUpdateInput(text=message)
        return StepApi(self._gql_client, self.auth).close_step(step_id, step_update)

    def add_iteration_step_artifact(self, step_id: int, step_artifacts: IterationStepArtifactInput) -> StepOutput:
        return StepApi(self._gql_client, self.auth).add_iteration_step_artifact(step_artifacts, step_id)

    def update_iteration_step_artifact(
        self,
        step_id: int,
        step_type: StepType,
        text: str | None = None,
        artifacts: list[IterationStepArtifactInput] | None = None,
    ) -> StepOutput:
        return StepApi(self._gql_client, self.auth).update_iteration_step_artifact(step_id, step_type, text, artifacts)

    def list_iterations(self, phase: int) -> list[IterationOutput]:
        return IterationApi(self._gql_client, self.auth).list_iterations(phase)

    def list_step_definitions(self, phase: int) -> list[StepOutput]:
        return PhaseApi(self._gql_client, self.auth).list_step_definitions(phase)

    def get_iteration(self, iteration_id: int) -> IterationOutput:
        return IterationApi(self._gql_client, self.auth).get_iteration(iteration_id)

    def get_iteration_by_index(self, phase_id: int, index: int) -> IterationOutput:
        return IterationApi(self._gql_client, self.auth).get_iteration_by_index(phase_id, index)

    def get_iteration_last_assets(self, iteration_id: int) -> IterationOutput:
        return IterationApi(self._gql_client, self.auth).get_iteration_last_assets(iteration_id)

    def create_iteration(self, phase_id: int) -> IterationOutput:
        return IterationApi(self._gql_client, self.auth).create_iteration(phase_id)

    def update_iteration(self, iteration_id: int, iteration: IterationInput) -> IterationOutput:
        return IterationApi(self._gql_client, self.auth).update_iteration(iteration, iteration_id)

    def delete_iteration(self, iteration_id: int) -> None:
        IterationApi(self._gql_client, self.auth).delete_iteration(iteration_id)

    def register_dataset_from_source(
        self,
        dataset: Dataset,
        project_id: int | None = None,
        phase_id: int | None = None,
        iteration_id: int | None = None,
        code_version_id: int | None = None,
    ) -> DatasetRegisterOutput:
        name = self.get_dataset_name(dataset)
        derived_from = self.get_dataset_derived_from(dataset)
        if isinstance(dataset.resource, tuple):
            # modeling dataset
            metadata_asdict = [data_source.metadata.asdict() for data_source in dataset.resource if data_source]
        else:
            metadata_asdict = [dataset.resource.metadata.asdict()]

        dataset_register_input = DatasetRegisterInput(
            name=name,
            type=dataset.type.value,
            datasetSources=metadata_asdict,
            inputs=derived_from,
            codeVersionId=code_version_id,
        )
        dataset_register_output = self.register_dataset(dataset_register_input, project_id, phase_id, iteration_id)
        dataset.latest_version_id = dataset_register_output["datasetVersion"]["id"]
        return dataset_register_output

    @staticmethod
    def get_dataset_name(dataset: Dataset) -> str:
        return f"dataset {datetime.time}" if dataset.name is None else dataset.name

    @staticmethod
    @deprecate(
        warn_at="23.1",
        fail_at="23.2",
        remove_at="23.3",
        reason="The 'get_dataset_inputs' method is renamed 'get_dataset_derived_from'. "
        "Using 'get_dataset_inputs' will raise an error in v{fail_at}. "
        "The method will be removed in v{remove_at}.",
    )
    def get_dataset_inputs(dataset: Dataset) -> list[int]:
        return Client.get_dataset_derived_from(dataset)

    @staticmethod
    def get_dataset_derived_from(dataset: Dataset) -> list[int]:
        return [] if dataset.derived_from is None else dataset.derived_from

    def register_dataset(
        self,
        dataset_register_input: DatasetRegisterInput,
        project_id: int | None = None,
        phase_id: int | None = None,
        iteration_id: int | None = None,
    ) -> DatasetRegisterOutput:
        try:
            data: DatasetRegisterOutput = GqlDatasetApi(self._gql_client, self.auth).register_dataset(
                dataset_register_input, project_id, phase_id, iteration_id
            )
            _logger.info(
                f"Successfully registered Dataset("
                f"name='{dataset_register_input.name}', "
                f"id={data['datasetVersion']['id']}, "
                f"version='{data['datasetVersion']['name']}', "
                f"type={dataset_register_input.type})."
            )
            return data
        except Exception as e:
            raise ValueError(f"The dataset register failed due to {e}") from e

    def get_project_and_workspace_references(
        self, project: str | int | None = None, workspace: str | int | None = None
    ):
        if project is None and self.project is not None:
            project = self.project.id
        if workspace is None and self.workspace is not None:
            workspace = self.workspace.id
        return project, workspace

    def get_project_and_workspace_refs_if_project_ref_is_str(
        self, project: str | int | None = None, workspace: str | int | None = None
    ):
        if project is None and self.project is not None:
            project = self.project.id
        elif isinstance(project, str):
            if workspace is None and self.workspace is not None:
                workspace = self.workspace.id
        return project, workspace

    def get_project_and_workspace_refs_if_project_ref_is_str_or_raise_error(
        self, project: str | int | None = None, workspace: str | int | None = None
    ) -> tuple[str | int, str | int | None]:
        project, workspace = self.get_project_and_workspace_refs_if_project_ref_is_str(project, workspace)
        if project is None:
            raise MissingReferenceError("project")
        return project, workspace

    def get_project_and_workspace_references_or_raise_error(
        self, project: str | int | None = None, workspace: str | int | None = None
    ) -> tuple[str | int, str | int | None]:
        project, workspace = self.get_project_and_workspace_references(project, workspace)
        if project is None:
            raise MissingReferenceError("project")
        return project, workspace

    @staticmethod
    def _get_model_name(library: str, technique: str, name: str | None = None) -> str:
        return name if name else f"{library} {technique} model"

    @deprecate(
        parameter="inputs",
        warn_at="23.1",
        fail_at="23.2",
        remove_at="23.3",
        reason="The 'inputs' parameter is renamed 'derived_from'. "
        "Using 'inputs' will raise an error in v{fail_at}. "
        "The parameter will be removed in v{remove_at}.",
    )
    def register_model(
        self,
        model: Model,
        project_id: int,
        phase_id: int | None = None,
        iteration_id: int | None = None,
        code_version_id: int | None = None,
        derived_from: list[int] | None = None,
        inputs: list[int] | None = None,
    ) -> ModelRegisterOutput:
        """Register a model.

        Parameters:
            model: The model to register
            project_id: The project ID
            phase_id: The phase ID
            iteration_id: The iteration ID
            code_version_id: The code version ID
            derived_from: The IDs of datasets the model is derived from
            inputs: Deprecated. Use `derived_from` instead

        Returns:
            The registered model.
        """
        if not derived_from and inputs:
            derived_from = inputs

        metrics = [vars(MetricInput(metric.key, metric.value)) for metric in model.metrics] if model.metrics else None
        properties = (
            [vars(PropertyInput(prop.key, prop.value)) for prop in model.properties] if model.properties else None
        )
        model_register_input = ModelRegisterInput(
            name=model.name,
            modelType=ModelType.OTHER.value,
            status=ModelVersionStatus.EXPERIMENTATION.value,
            inputs=derived_from or [],
            metrics=metrics,
            properties=properties,
            algorithmName=model.technique,
            framework=model.library,
            codeVersionId=code_version_id,
        )
        return GqlModelApi(self._gql_client, self.auth).register_model(
            model_register_input, project_id, phase_id, iteration_id
        )

    def get_last_assets(self, target_types: list[str], page):
        return LastAssetApi(self._gql_client, self.auth).get_last_assets(target_types, page)

    def get_user_and_default_workspace(self):
        return UserAndDefaultWorkspaceApi(self._gql_client, self.auth).get_user_and_default_workspace()

    def create_code_gql(self, project_id: int, code: CodeInput):
        return GqlCodeApi(self._gql_client, self.auth).create_code(project_id, code)

    def create_code_version_gql(self, code_id: int, code_version: CodeVersionCreateBody):
        return GqlCodeVersionApi(self._gql_client, self.auth).create_code_version(code_id, code_version)

    def get_code(self, code: str | int, project_id: int | None = None):
        if project_id is None:
            raise MissingReferenceError("project")
        return GqlCodeApi(self._gql_client, self.auth).get_code(code, project_id)

    def get_code_version(self, code_version: str | int, code_id: int | None = None):
        if code_id is None:
            raise MissingReferenceError("code")
        return GqlCodeVersionApi(self._gql_client, self.auth).get_code_version(code_version, code_id)



from __future__ import annotations

import logging
from io import IOBase

from PIL import Image

from vectice.api.json.iteration import IterationStepArtifactInput
from vectice.api.json.step import StepType
from vectice.models.step import Step
from vectice.utils.common_utils import _check_read_only, _get_image_variables

_logger = logging.getLogger(__name__)


class StepImage(Step):
    """Model a Vectice step's image.

    A StepImage stores an image.
    """

    def __init__(self, step: Step, image: str | IOBase | None = None):
        """Initialize a step's image.

        Parameters:
            step: The step
            image: The step's image.
        """
        super().__init__(step.id, step.iteration, step.name, step.index, step._description, artifacts=step.artifacts)
        self._image: str | IOBase | None = image

    def __repr__(self):
        return f"StepImage(name='{self.name}', id={self.id}, description='{self._description}', image={self.image})"

    @property
    def image(self) -> str | IOBase | None:
        """The step's image.

        Returns:
            The step's image.
        """
        return self._image

    @image.setter
    def image(self, value: str | IOBase | Image.Image):
        """The step's image.

        Typical usage example with image path:

        ```python
        my_iteration = my_phase.iteration()
        my_iteration.step_explanation = "accuracy.png"
        ```

        Typical usage example with object in memory:

        ```python
        my_iteration = my_phase.iteration()
        with open("my_image.png", "rb") as image:
            my_iteration.step_explanation = image
        ```

        Typical usage example with PIL object in memory:

        ```python
        my_iteration = my_phase.iteration()
        with Image.open("my_image.png") as image:
            my_iteration.step_explanation = image
        ```

        Parameters:
            value: The image path or object.
        """
        _check_read_only(self.iteration)
        image, filename = _get_image_variables(value)
        self._image = filename
        attachments_output = self._client.create_phase_attachments(
            [("file", (filename, image))], self.phase.id, self.project.id
        )
        step_artifacts = [
            IterationStepArtifactInput(id=artifact["entityId"], type="EntityFile") for artifact in attachments_output
        ]
        self._client.update_iteration_step_artifact(
            self.id,
            StepType.StepImage,
            self.text,
            step_artifacts,
        )



from __future__ import annotations

from vectice.models.resource.metadata.base import (
    DatasetSourceOrigin,
    DatasetSourceType,
    DatasetSourceUsage,
    DatasetType,
    Metadata,
)
from vectice.models.resource.metadata.db_metadata import Column, DBMetadata, MetadataDB
from vectice.models.resource.metadata.files_metadata import File, FilesMetadata
from vectice.utils.deprecation import deprecate

__all__ = [
    "DBMetadata",
    "Column",
    "MetadataDB",
    "DatasetSourceOrigin",
    "DatasetSourceUsage",
    "DatasetSourceType",
    "DatasetType",
    "File",
    "FilesMetadata",
    "Metadata",
]


@deprecate(
    warn_at="23.1",
    fail_at="23.3",
    remove_at="23.4",
    reason="The `vectice.models.datasource.datawrapper.metadata` module "
    "is deprecated in favor the `vectice.models.resource.metadata` module. "
    "Importing from the deprecated module will fail in v{fail_at}, "
    "and the module will be removed in v{remove_at}.",
)
def _warn() -> None:
    pass


_warn()



from __future__ import annotations

from typing import Any

from vectice.models.resource.metadata.base import DatasetSourceOrigin, DatasetSourceType, DatasetType
from vectice.utils.deprecation import deprecate


def __getattr__(name: str) -> Any:
    if name == "SourceUsage":
        return _get_deprecated_source_usage()
    if name == "SourceOrigin":
        return _get_deprecated_source_origin()
    if name == "SourceType":
        return _get_deprecated_source_type()

    raise AttributeError(f"'{__name__}' module has no attribute '{name}'")


@deprecate(
    warn_at="23.1",
    fail_at="23.3",
    remove_at="23.4",
    reason="The `SourceUsage` enumeration is renamed `DatasetType`. "
    "Importing and using `SourceUsage` will fail in v{fail_at}. "
    "The enumeration will be removed in v{remove_at}.",
)
def _get_deprecated_source_usage():
    return DatasetType


@deprecate(
    warn_at="23.1",
    fail_at="23.3",
    remove_at="23.4",
    reason="The `SourceOrigin` enumeration is renamed `DatasetSourceOrigin`. "
    "Importing and using `SourceOrigin` will fail in v{fail_at}. "
    "The enumeration will be removed in v{remove_at}.",
)
def _get_deprecated_source_origin():
    return DatasetSourceOrigin


@deprecate(
    warn_at="23.1",
    fail_at="23.3",
    remove_at="23.4",
    reason="The `SourceType` enumeration is renamed `DatasetSourceType`. "
    "Importing and using `SourceType` will fail in v{fail_at}. "
    "The enumeration will be removed in v{remove_at}.",
)
def _get_deprecated_source_type():
    return DatasetSourceType


@deprecate(
    warn_at="23.1",
    fail_at="23.3",
    remove_at="23.4",
    reason="The `vectice.models.datasource.datawrapper.metadata.metadata` module "
    "is deprecated in favor the `vectice.models.resource.metadata.base` module. "
    "Importing from the deprecated module will fail in v{fail_at}, "
    "and the module will be removed in v{remove_at}.",
)
def _warn() -> None:
    pass


_warn()



from __future__ import annotations

from typing import TYPE_CHECKING

from vectice.api.json.iteration import IterationStepArtifact, IterationStepArtifactInput, IterationStepArtifactType
from vectice.models.dataset import Dataset
from vectice.models.datasource.datawrapper import FileDataWrapper, GcsDataWrapper, S3DataWrapper
from vectice.models.datasource.datawrapper.data_wrapper import DataWrapper
from vectice.models.resource import FileResource, GCSResource, S3Resource
from vectice.models.resource.metadata import DatasetSourceUsage, DatasetType
from vectice.utils.deprecation import ignore_deprecation_warnings

if TYPE_CHECKING:
    from logging import Logger

    from vectice.models import Step
    from vectice.models.resource.base import Resource


def existing_dataset_logger(data: dict, dataset_name: str, _logger: Logger):
    if data["useExistingDataset"]:
        existing_dataset_version = f"Dataset: {dataset_name}"
        if data["useExistingVersion"]:
            existing_dataset_version += f" with Version: {data['datasetVersion']['name']}"
        _logger.debug(f"{existing_dataset_version} already exists.")


def existing_model_logger(data: dict, model_name: str, _logger: Logger):
    if data["useExistingModel"]:
        _logger.info(f"Model: {model_name} already exists.")


def link_dataset_to_step(
    step_artifact: IterationStepArtifactInput,
    dataset: Dataset,
    data: dict,
    _logger: Logger,
    step: Step,
):
    """Link a dataset to a step.

    1. We get the current phase if it exists
        - No phase results in a warning / failure (we can't assume phases are ordinal so this fails/exits)
        - We need a phase to be called at some point at the very least
    2. We get the current iteration if it exists (e.g. a user did `phase.iteration()`)
        - No iteration results in us getting one/creating one.
    3. We get the current step if it exists (e.g. a user did `iteration.step()`)
    3. If there is no current step
        - We check if the iteration has any steps that are not completed

    Parameters:
        step_artifact: The artifact to link.
        dataset: The dataset.
        data: The artifact metadata.
        step: The step to link to.
    """
    client = step._client
    step_artifacts = get_updated_step_artifacts(step, _logger)
    dataset_type_name = dataset.type.name
    matching_datasets = [
        artifact
        for artifact in step_artifacts
        if artifact.type == IterationStepArtifactType.DataSetVersion and artifact.dataset_version_id == step_artifact.id
    ]
    if datasets_are_new_in_step(step_artifacts, matching_datasets):
        step_output = client.add_iteration_step_artifact(step.id, step_artifact)
        _logger.info(
            f"Successfully added Dataset(name='{dataset.name}', id={data['datasetVersion']['id']}, "
            f"version='{data['datasetVersion']['name']}', type={dataset_type_name}) to {step_output.name}"
        )
    elif datasets_already_exist_in_step(step_artifacts, matching_datasets):
        _logger.info(
            f"The step {step.name} already has Dataset(name='{dataset.name}', id={data['datasetVersion']['id']}, "
            f"version='{data['datasetVersion']['name']}', type={dataset_type_name}) linked."
        )
    else:
        _logger.warning(
            f"There are no active steps to attach the registered Dataset(name='{dataset.name}', id={data['datasetVersion']['id']}, "
            f"version='{data['datasetVersion']['name']}', type={dataset_type_name}) to."
        )


def get_updated_step_artifacts(step: Step, _logger: Logger):
    iteration = step.iteration
    refresh_step = iteration.step(step.name)
    with ignore_deprecation_warnings():
        if not refresh_step.completed:
            return refresh_step.artifacts
    return []


def datasets_are_new_in_step(
    step_artifacts: list[IterationStepArtifact], matching_datasets: list[IterationStepArtifact]
):
    return len(step_artifacts) == 0 or (len(step_artifacts) >= 1 and not any(matching_datasets))


def datasets_already_exist_in_step(
    step_artifacts: list[IterationStepArtifact], matching_datasets: list[IterationStepArtifact]
):
    return len(step_artifacts) >= 1 and any(matching_datasets)


def link_assets_to_step(
    step: Step,
    step_artifact: IterationStepArtifactInput,
    name: str,
    data: dict,
    _logger: Logger,
    attachments: list[IterationStepArtifactInput] | None = None,
):
    """Link assets to a step.

    The step is refreshed, the step status is checked and the link and logger is used for the relevant artifact.

    Parameters:
        step: The current step.
        step_artifact: A step artifact to link.
        name: The resulting dataset name.
        data: The dataset metadata dictionary.
        attachments: Additional artifact to link.

    Raises:
        RuntimeError: When the step artifact type is neither `DataSetVersion` or `ModelVersion`.
    """
    client = step._client
    refresh_step = step._iteration.step(step.name)
    with ignore_deprecation_warnings():
        refresh_step_completed = refresh_step and refresh_step.completed
    if refresh_step and not refresh_step_completed:
        step_id, step_name = refresh_step.id, refresh_step.name
        step_artifacts = refresh_step.artifacts
    else:
        raise RuntimeError(f"The step: {step.name} is not active!")
    if step_artifact.type == IterationStepArtifactType.DataSetVersion.value:
        _dataset_link_and_logger(client, step_artifact, step_artifacts, name, data, step_name, step_id, _logger)
    elif step_artifact.type == IterationStepArtifactType.ModelVersion.value:
        _model_link_and_logger(client, step_artifact, step_artifacts, name, data, step_name, step_id, _logger)
    else:
        raise RuntimeError("Vectice Error: The step artifact type was not set.")
    if attachments:
        _attachment_link_and_logger(client, attachments, step_artifacts, name, data, step_name, step_id, _logger)


def _dataset_link_and_logger(
    client,
    step_artifact: IterationStepArtifactInput,
    step_artifacts: list[IterationStepArtifact],
    name: str,
    data: dict,
    step_name: str,
    step_id: int,
    _logger: Logger,
):
    check_assets = [artifact for artifact in step_artifacts if artifact.type == IterationStepArtifactType.ModelVersion]
    match_dataset_assets = [
        artifact
        for artifact in step_artifacts
        if artifact.type == IterationStepArtifactType.DataSetVersion and artifact.dataset_version_id == step_artifact.id
    ]
    if (len(step_artifacts) == 0 or len(step_artifacts) == len(check_assets)) or (
        len(step_artifacts) >= 1 and not any(match_dataset_assets)
    ):
        step_output = client.add_iteration_step_artifact(step_id, step_artifact)
        _logger.info(
            f"Successfully added Dataset(name='{name}', id={data['datasetVersion']['id']}, version='{data['datasetVersion']['name']}', type=MODELING) to {step_output.name}"
        )
    elif len(step_artifacts) >= 1 and any(match_dataset_assets):
        _logger.info(
            f"The step {step_name} already has Dataset(name='{name}', id={data['datasetVersion']['id']}, version='{data['datasetVersion']['name']}', type=MODELING) linked."
        )
    else:
        raise RuntimeError("Vectice Error: The dataset link failed")


def _model_link_and_logger(
    client,
    step_artifact: IterationStepArtifactInput,
    step_artifacts: list[IterationStepArtifact],
    name: str,
    data: dict,
    step_name: str,
    step_id: int,
    _logger: Logger,
):
    check_assets = [
        artifact for artifact in step_artifacts if artifact.type == IterationStepArtifactType.DataSetVersion
    ]
    match_model_assets = [
        artifact
        for artifact in step_artifacts
        if artifact.type == IterationStepArtifactType.ModelVersion and artifact.model_version_id == step_artifact.id
    ]
    if (len(step_artifacts) == 0 or len(step_artifacts) == len(check_assets)) or (
        len(step_artifacts) >= 1 and not any(match_model_assets)
    ):
        step_output = client.add_iteration_step_artifact(step_id, step_artifact)
        _logger.info(
            f"Successfully added Model(name='{name}', version='{data['modelVersion']['name']}') to {step_output.name}"
        )
    elif len(step_artifacts) >= 1 and any(match_model_assets):
        _logger.info(
            f"The step {step_name} already has Model(name='{name}', version='{data['modelVersion']['name']}') linked."
        )
    else:
        raise RuntimeError("Vectice Error: The model link failed")


def _attachment_link_and_logger(
    client,
    attachments: list[IterationStepArtifactInput],
    step_artifacts: list[IterationStepArtifact],
    name: str,
    data: dict,
    step_name: str,
    step_id: int,
    _logger: Logger,
):
    check_assets = [artifact for artifact in step_artifacts if artifact.type == IterationStepArtifactType.EntityFile]
    for attachment in attachments:
        match_attachment_assets = [
            artifact
            for artifact in step_artifacts
            if artifact.type == IterationStepArtifactType.EntityFile
            and artifact.entity_file_id == attachment.entity_file_id
        ]
        if (len(step_artifacts) == 0 or len(step_artifacts) == len(check_assets)) or (
            len(step_artifacts) >= 1 and not any(match_attachment_assets)
        ):
            attachment.pop("entityFileId")
            step_output = client.add_iteration_step_artifact(step_id, attachment)
            _logger.info(
                f"Successfully added Attachment: {attachment.id} from Model(name='{name}', version='{data['modelVersion']['name']}') to {step_output.name}"
            )
        elif len(step_artifacts) >= 1 and any(match_attachment_assets):
            _logger.info(
                f"The step {step_name} already has Attachment: {attachment.id} from Model(name='{name}', version='{data['modelVersion']['name']}') linked."
            )
        else:
            raise RuntimeError("Vectice Error: The attachment link failed")


def wrapper_to_resource(wrapper: DataWrapper) -> Resource:
    if isinstance(wrapper, FileDataWrapper):
        return FileResource(path=wrapper.path)
    elif isinstance(wrapper, GcsDataWrapper):
        return GCSResource(
            gcs_client=wrapper.gcs_client,
            bucket_name=wrapper.bucket_name,
            resource_paths=wrapper.resource_paths,
        )
    elif isinstance(wrapper, S3DataWrapper):
        return S3Resource(
            s3_client=wrapper.s3_client,
            bucket_name=wrapper.bucket_name,
            resource_path=wrapper.resource_path,
        )
    raise ValueError(f"Unhandled wrapper type: {wrapper.__class__}")


def wrapper_to_dataset(wrapper: DataWrapper | tuple[DataWrapper, ...]) -> Dataset:
    if isinstance(wrapper, DataWrapper):
        return Dataset(
            type=wrapper.type,
            name=wrapper.name,
            resource=wrapper_to_resource(wrapper),
            derived_from=wrapper.derived_from,  # type: ignore[arg-type]
        )

    derived_from = []
    training_resource = testing_resource = validation_resource = None
    for wrap in wrapper:
        if wrap.usage is DatasetSourceUsage.TRAINING:
            training_resource = wrapper_to_resource(wrap)
        elif wrap.usage is DatasetSourceUsage.TESTING:
            testing_resource = wrapper_to_resource(wrap)
        elif wrap.usage is DatasetSourceUsage.VALIDATION:
            validation_resource = wrapper_to_resource(wrap)
        if wrap.derived_from:
            derived_from.extend(wrap.derived_from)
    return Dataset(
        type=DatasetType.MODELING,
        name=wrapper[0].name,
        training_resource=training_resource,
        testing_resource=testing_resource,
        validation_resource=validation_resource,
        derived_from=derived_from,  # type: ignore[arg-type]
    )



from __future__ import annotations

import logging
import os
from contextlib import suppress
from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING

import dotenv
from rich.table import Table

from vectice.api import Client
from vectice.api.json.last_assets import ActivityTargetType
from vectice.models.workspace import Workspace
from vectice.utils.common_utils import _temp_print, hide_logs
from vectice.utils.configuration import Configuration
from vectice.utils.last_assets import _connection_logging, _get_last_assets, _get_last_user_and_default_workspace
from vectice.utils.logging_utils import CONNECTION_PROJECT_LOGGING, CONNECTION_WORKSPACE_LOGGING, format_description

if TYPE_CHECKING:
    from vectice.models.project import Project


_logger = logging.getLogger(__name__)
DEFAULT_API_ENDPOINT = "https://app.vectice.com"
CAN_NOT_BE_EMPTY_ERROR_MESSAGE = "%s can not be empty."


class Connection:
    """Connect to the Vectice backend (application).

    The Connection class encapsulates a connection to the Vectice App.
    Thus, it authenticates and connects to Vectice.
    This allows you to start interacting with your Vectice assets.

    A Connection can be initialized in three ways:

    1. Passing the relevant arguments to authenticate and connect to Vectice:

        ```python
        import vectice

        connection = vectice.connect(
            api_token="API_TOKEN_FROM_VECTICE",
            host="https://app.vectice.com",
        )
        ```

    2. Passing the path to a configuration file:

        ```python
        import vectice

        connection = vectice.connect(config="vectice_config.json")
        ```

    3. Letting Vectice find the configuration file in specific locations:

        ```python
        import vectice

        connection = vectice.connect()
        ```

    See [`Connection.connect`][vectice.connection.Connection.connect] for more info.
    """

    USER_FILES_PATH = [
        ".vectice",
        str(Path.home() / ".vectice"),
        ".env",
        str(Path.home() / ".env"),
        "/etc/vectice/api.cfg",
    ]

    def __init__(
        self,
        api_token: str,
        host: str,
        workspace: str | int | None = None,
        project: str | int | None = None,
    ):
        """Initialize a connection.

        Parameters:
            api_token: Your private api token.
            host: The address of the Vectice application.
            workspace: The workspace you want to work in.
            project: The project you want to work in.

        Raises:
            RuntimeError: When the API and backend versions are incompatible.
        """
        logging.getLogger("Client").propagate = True
        self._client = Client(
            workspace=workspace,
            project=project,
            token=api_token,
            api_endpoint=host,
            auto_connect=True,
            allow_self_certificate=True,
        )
        compatibility = self._client.check_compatibility()
        if compatibility.status != "OK":
            if compatibility.status == "Error":
                _logger.error(f"compatibility error: {compatibility.message}")
                raise RuntimeError(f"compatibility error: {compatibility.message}")
            else:
                _logger.warning(f"compatibility warning: {compatibility.message}")

    def __repr__(self) -> str:
        return (
            "Connection("
            + f"workspace={self._client.workspace.name if self._client.workspace else 'None'}, "
            + f"host={self._client.auth.api_base_url}, "
        )

    @property
    def version_api(self) -> str:
        return self._client.version_api

    @property
    def version_backend(self) -> str:
        return self._client.version_backend

    def workspace(self, workspace: str | int) -> Workspace:
        """Get a workspace.

        Parameters:
            workspace: The id or the name of the desired workspace.

        Returns:
            The desired workspace.
        """
        output = self._client.get_workspace(workspace)
        result = Workspace(output.id, output.name, output.description)
        result.__post_init__(self._client, self)
        return result

    def list_workspaces(self) -> None:
        """List the workspaces to which this connection has access and log the workspaces in a tabular format.

        Returns:
            The workspaces to which this connection has access.
        """
        workspace_outputs = self._client.list_workspaces().list
        user_name, _ = _get_last_user_and_default_workspace(self._client)

        rich_table = Table(expand=True, show_edge=False)

        rich_table.add_column("workspace id", justify="left", no_wrap=True, min_width=4, max_width=10)
        rich_table.add_column("name", justify="left", no_wrap=True, max_width=20)
        rich_table.add_column("description", justify="left", no_wrap=True, max_width=50)

        for count, workspace in enumerate(workspace_outputs, 1):
            if count > 10:
                break
            rich_table.add_row(str(workspace.id), workspace.name, format_description(workspace.description))

        description = f"""There are {len(workspace_outputs)} workspaces for user {user_name}."""
        tips = dedent(
            """
        >> To access your personal workspace, use connection.my_workspace or connect.workspaces[0]
        >> To access a specific workspace, use connection.workspace(Workspace ID)"""
        ).lstrip()
        link = dedent(
            f"""
        For quick access to the list of workspaces in the Vectice web app, visit:
        {self._client.auth._API_BASE_URL}/workspaces"""
        ).lstrip()
        _temp_print(description)
        _temp_print(table=rich_table)
        _temp_print(tips)
        _temp_print(link)

    @property
    def workspaces(self) -> list[Workspace] | None:
        """List the workspaces to which this connection has access.

        Returns:
            The workspaces to which this connection has access.
        """
        _, default_workspace_id = _get_last_user_and_default_workspace(self._client)
        outputs = self._client.list_workspaces()
        results: list[Workspace] = []
        for output in outputs.list:
            workspace = Workspace(id=output.id, name=output.name, description=output.description)
            workspace.__post_init__(self._client, self)
            if output.id == default_workspace_id:
                results.insert(0, workspace)
                continue
            results.append(workspace)
        return results

    @property
    def my_workspace(self) -> Workspace:
        """Retrieve your personal workspace.

        Returns:
            Personal workspace.
        """
        asset = self._client.get_user_and_default_workspace()
        if not asset.get("defaultWorkspace"):
            raise ValueError("Default workspace is not set.")
        return self.workspace(int(asset["defaultWorkspace"]["id"]))

    @property
    def last_workspace(self) -> Workspace:
        """Retrieve last workspace with activity.

        Returns:
            Last workspace with activity.
        """
        target_types = [activity for activity in ActivityTargetType]
        asset = _get_last_assets(target_types, self._client, _logger)
        if not asset or not asset.get("workspace"):
            raise ValueError("Workspace with activity not found.")
        return self.workspace(int(asset["workspace"]["id"]))

    @staticmethod
    def connect(
        api_token: str | None = None,
        host: str | None = None,
        config: str | None = None,
        workspace: str | int | None = None,
        project: str | None = None,
    ) -> Connection | Workspace | Project | None:
        """Method to connect to the Vectice backend (application).

        Authentication credentials are retrieved, in order, from:

        1. keyword arguments
        2. configuration file (`config` parameter)
        3. environment variables
        4. environment files in the following order
            - `.vectice` of the working directory
            - `.vectice` of the user home directory
            - `.env` of the working directory
            - `.env` of the user home directory
            - `/etc/vectice/api.cfg` file

        This method uses the `api_token`, `host`, `workspace`, `project` arguments
        or the JSON config provided. The JSON config file is available from the Vectice
        webapp when creating an API token.

        Parameters:
            api_token: The api token provided by the Vectice webapp.
            host: The backend host to which the client will connect.
                If not found, the default endpoint https://app.vectice.com is used.
            config: A JSON config file containing keys VECTICE_API_TOKEN and
                VECTICE_API_ENDPOINT as well as optionally WORKSPACE and PROJECT.
            workspace: The name of an optional workspace to return.
            project: The name of an optional project to return.

        Raises:
            ValueError: When a project is specified without a workspace.

        Returns:
            A Connection, Workspace, or Project.
        """
        host = host or Connection._get_host(config)
        api_token = api_token or Connection._get_api_token(host, config)
        workspace = workspace or Connection._get_workspace(config)
        project = project or Connection._get_project(config)
        connection = Connection(api_token=api_token, host=host, workspace=workspace, project=project)
        user_name, workspace_id = _get_last_user_and_default_workspace(connection._client)
        url = connection._client.auth.api_base_url
        if workspace and not project:
            return connection.get_workspace(workspace, user_name, url)
        if workspace and project:
            return connection.get_project(workspace, project, user_name, url)
        _connection_logging(_logger, user_name, url, workspace_id)
        return connection

    @staticmethod
    def _get_host(config: str | None) -> str:
        try:
            return Connection._get_config_item("VECTICE_API_ENDPOINT", config)  # type: ignore[return-value]
        except ValueError:
            _logger.debug(f"No VECTICE_API_ENDPOINT provided. Using default endpoint {DEFAULT_API_ENDPOINT}")
            return DEFAULT_API_ENDPOINT

    @staticmethod
    def _get_api_token(host: str | None, config: str | None) -> str:
        try:
            return Connection._get_config_item("VECTICE_API_TOKEN", config)  # type: ignore[return-value]
        except ValueError as error:
            raise ValueError(
                f"You must provide the api_token. You can generate them by going to the page {host}/account/api-keys"
            ) from error

    @staticmethod
    def _get_workspace(config) -> str | int | None:
        try:
            return Connection._get_config_item("WORKSPACE", config)
        except ValueError:
            return None

    @staticmethod
    def _get_project(config) -> str | None:
        try:
            return str(Connection._get_config_item("PROJECT", config))
        except ValueError:
            return None

    @staticmethod
    def _get_config_item(item_name: str, config_path: str | None) -> str | int:
        # search in provided config file
        with suppress(KeyError, SyntaxError, TypeError):
            config = Configuration(config_path)  # type: ignore[arg-type]
            if config[item_name]:
                _logger.debug(f"Found {item_name} in {config_path}")
                return config[item_name]
        # search in environment variables
        item = os.environ.get(item_name)
        if item:
            _logger.debug(f"Found {item_name} in environment variables.")
            return item
        # search in user configuration files
        for path in Connection.USER_FILES_PATH:
            with hide_logs("dotenv"):
                item = dotenv.get_key(path, item_name)
            if item:
                _logger.debug(f"Found {item_name} in {path}")
                return item
        raise ValueError(f"Could not find {item_name} in user configuration")

    def get_workspace(self, workspace: int | str, user_name: str, host: str) -> Workspace:
        workspace_output: Workspace = self.workspace(workspace)
        _logger.info(
            CONNECTION_WORKSPACE_LOGGING.format(
                user=user_name, workspace_name=workspace_output.name, url=host, workspace_id=workspace_output.id
            )
        )
        return workspace_output

    def get_project(self, workspace: int | str, project: str, user_name: str, host: str) -> Project:
        workspace_output = self.workspace(workspace)
        project_output: Project = workspace_output.project(project)
        _logger.info(
            CONNECTION_PROJECT_LOGGING.format(
                user=user_name,
                project_name=project_output.name,
                url=host,
                workspace_id=workspace_output.id,
                project_id=project_output.id,
            )
        )
        return project_output



from __future__ import annotations

import logging
from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING

from vectice.models.resource.metadata import DatasetType
from vectice.utils.deprecation import deprecate

if TYPE_CHECKING:
    from vectice.models.resource.metadata import DatasetSourceUsage, FilesMetadata

_logger = logging.getLogger(__name__)


class DataWrapper(metaclass=ABCMeta):
    """Deprecated. Base class for DataWrapper.

    WARNING: **Wrappers are deprecated.**
    Instead, use [`Dataset`][vectice.models.dataset.Dataset]
    and [`Resource`][vectice.models.resource.base.Resource].

    Use DataWrapper subclasses to assign datasets to steps.  The
    Vectice library supports a handful of common cases.  Additional
    cases are generally easy to supply by deriving from this base
    class.  In particular, subclasses must override this class'
    abstact methods (`_build_metadata()`, `_fetch_data()`).

    Examples:
        To create a custom data wrapper, inherit from `DataWrapper`,
        and implement the `_build_metadata()` and `_fetch_data()` methods:

        ```python
        from vectice.models.datasource.datawrapper import DataWrapper
        from vectice.models.datasource.datawrapper.metadata import FilesMetadata

        class MyDataWrapper(DataWrapper):
            _origin = "Data source name"

            def _build_metadata(self) -> FilesMetadata:  # (1)
                files = ...  # fetch file list from your custom storage
                total_size = ...  # compute total file size
                return FilesMetadata(
                    size=total_size,
                    origin=self._origin,
                    files=files,
                    usage=self.usage,
                )

            def _fetch_data(self) -> dict[str, bytes]:
                files_data = {}
                for file in self.metadata.files:
                    file_contents = ...  # fetch file contents from your custom storage
                    files_data[file.name] = file_contents
                return files_data
        ```

        1. Return [FilesMetadata][vectice.models.datasource.datawrapper.metadata.FilesMetadata] for data stored in files,
            [DBMetadata][vectice.models.datasource.datawrapper.metadata.DBMetadata] for data stored in a database.
    """

    @abstractmethod
    @deprecate(
        warn_at="23.1",
        fail_at="23.3",
        remove_at="23.4",
        reason="The data wrappers are deprecated in favor or "
        "the new Dataset and Resource classes. "
        "Using data wrappers will fail in v{fail_at}. "
        "They will be removed in v{remove_at}.",
    )
    @deprecate(
        parameter="inputs",
        warn_at="23.1",
        fail_at="23.2",
        remove_at="23.3",
        reason="The 'inputs' parameter is renamed 'derived_from'. "
        "Using 'inputs' will raise an error in v{fail_at}. "
        "The parameter will be removed in v{remove_at}.",
    )
    def __init__(
        self,
        name: str,
        usage: DatasetSourceUsage | None = None,
        derived_from: list[int] | None = None,
        inputs: list[int] | None = None,
        type: DatasetType | None = None,
    ):
        """Initialize a data wrapper.

        Parameters:
            name: The name of the data wrapper.
            usage: The usage of the dataset.
            derived_from: The list of dataset ids to create a new dataset from.
            inputs: Deprecated. Use `derived_from` instead.
            type: The type of the dataset.

        Raises:
            ValueError: When both `type` and `value` arguments are `None`.
        """
        if not derived_from and inputs:
            derived_from = inputs

        if type is None:
            if usage is None:
                raise ValueError("'type' and 'usage' cannot be both None")
            type = DatasetType.MODELING

        elif type is DatasetType.MODELING and usage is None:
            raise ValueError(f"'usage' must be set when 'type' is {DatasetType.MODELING!r}")

        self._old_name = name
        self._name = name
        self._type = type
        self._usage = usage
        self._derived_from = derived_from
        self._metadata = None
        self._data = None

    @property
    def data(self) -> dict[str, bytes]:
        """The wrapper's data.

        Returns:
            The wrapper's data.
        """
        if self._data is None:
            self._data = self._fetch_data()  # type: ignore[assignment]
        return self._data  # type: ignore[return-value]

    @abstractmethod
    def _fetch_data(self) -> dict[str, bytes]:
        pass

    @abstractmethod
    def _build_metadata(self) -> FilesMetadata:
        pass

    @property
    def name(self) -> str:
        """The wrapper's name.

        Returns:
            The wrapper's name.
        """
        return self._name

    @name.setter
    def name(self, value):
        """Set the wrapper's name.

        Parameters:
            value: The wrapper's name.
        """
        self._name = value
        self._clear_data_and_metadata()

    @property
    def type(self) -> DatasetType:
        """The wrapper's dataset type.

        Returns:
            The wrapper's dataset type.
        """
        return self._type

    @property
    def usage(self) -> DatasetSourceUsage | None:
        """The wrapper's dataset usage.

        Returns:
            The wrapper's dataset usage.
        """
        return self._usage

    @property
    @deprecate(
        warn_at="23.1",
        fail_at="23.2",
        remove_at="23.3",
        reason="The 'inputs' property is renamed 'derived_from'. "
        "Using 'inputs' will raise an error in v{fail_at}. "
        "The property will be removed in v{remove_at}.",
    )
    def inputs(self) -> list[int] | None:
        """Deprecated. Use `derived_from` instead.

        Returns:
            The datasets from which this wrapper is derived.
        """
        return self.derived_from

    @property
    def derived_from(self) -> list[int] | None:
        """The datasets from which this wrapper is derived.

        Returns:
            The datasets from which this wrapper is derived.
        """
        return self._derived_from

    @property
    def metadata(self) -> FilesMetadata:
        """The wrapper's metadata.

        Returns:
            The wrapper's metadata.
        """
        if self._metadata is None:
            self.metadata = self._build_metadata()
        return self._metadata  # type: ignore[return-value]

    @metadata.setter
    def metadata(self, value):
        """Set the wrapper's metadata.

        Parameters:
            value: The metadata to set.
        """
        self._metadata = value

    def _clear_data_and_metadata(self):
        self._data = None
        self._metadata = None



from __future__ import annotations

from typing import TYPE_CHECKING

from vectice.models.resource.base import Resource
from vectice.models.resource.metadata import DatasetSourceOrigin
from vectice.models.resource.metadata.files_metadata import File, FilesMetadata

if TYPE_CHECKING:
    from google.cloud.storage import Blob, Bucket, Client


class GCSResource(Resource):
    """Wrap columnar data and its metadata in GCS.

    Vectice stores metadata -- data about your dataset -- communicated
    with a resource.  Your actual dataset is not stored by Vectice.

    This resource wraps data that you have stored in Google Cloud
    Storage.  You assign it to a step.

    ```python
    from vectice import GCSResource
    from google.cloud.storage import Client

    my_service_account_file = "MY_SERVICE_ACCOUNT_JSON_PATH"  # (1)
    gcs_client = Client.from_service_account_json(json_credentials_path=my_service_account_file)  # (2)
    gcs_resource = GCSResource(
        gcs_client,
        bucket_name="my_bucket",
        resource_paths="my_folder/my_filename",
    )
    ```

    1. See [Service account credentials](https://developers.google.com/workspace/guides/create-credentials#service-account).
    1. See [GCS docs](https://cloud.google.com/python/docs/reference/storage/latest/modules).

    Note that these three concepts are distinct, even if easily conflated:

    * Where the data is stored
    * The format at rest (in storage)
    * The format when loaded in a running Python program

    Notably, the statistics collectors provided by Vectice operate
    only on this last and only in the event that the data is loaded as
    a pandas dataframe.
    """

    _origin = DatasetSourceOrigin.GCS.value

    def __init__(
        self,
        gcs_client: Client,
        bucket_name: str,
        resource_paths: str | list[str],
    ):
        """Initialize a GCS resource.

        Parameters:
            gcs_client: The `google.cloud.storage.Client` used
                to interact with Google Cloud Storage.
            bucket_name: The name of the bucket to get data from.
            resource_paths: The paths of the resources to get.
        """
        super().__init__()
        self.bucket_name = bucket_name
        self.resource_paths = resource_paths if isinstance(resource_paths, list) else [resource_paths]
        self.gcs_client = gcs_client

    def _fetch_data(self) -> dict[str, bytes]:
        datas = {}
        for path in self.resource_paths:
            blob = self._get_blob(path)
            datas[f"{self.bucket_name}/{path}"] = blob.download_as_bytes()
        return datas

    def _build_metadata(self) -> FilesMetadata:
        files = []
        size = 0
        for path in self.resource_paths:
            blob = self._get_blob(path)
            blob_file = self._build_file_from_blob(blob)
            files.append(blob_file)
            size += blob_file.size
        metadata = FilesMetadata(
            size=size,
            origin=self._origin,
            files=files,
        )
        return metadata

    def _get_blob(self, path: str) -> Blob:
        from google.cloud import storage

        bucket: Bucket = storage.Bucket(self.gcs_client, name=self.bucket_name)
        blob = bucket.get_blob(blob_name=path)
        if blob is None:
            raise NoSuchGCSResourceError(self.bucket_name, path)
        blob.reload()
        return blob

    def _build_file_from_blob(self, blob: Blob) -> File:
        return File(
            name=blob.name,
            size=blob.size,
            fingerprint=blob.md5_hash,
            created_date=blob.time_created.isoformat(),
            updated_date=blob.updated.isoformat(),
            uri=f"gs://{self.bucket_name}/{blob.name}",
        )


class NoSuchGCSResourceError(Exception):
    def __init__(self, bucket: str, resource: str):
        self.message = f"{resource} does not exist in the GCS bucket {bucket}."
        super().__init__(self.message)



from __future__ import annotations

import urllib.parse

from vectice.api.http_error_handlers import InvalidReferenceError, MissingReferenceError
from vectice.api.json import PagedResponse, ProjectInput, ProjectOutput
from vectice.api.rest_api import HttpError, RestApi
from vectice.api.workspace import WorkspaceApi

INVALID_PROJECT_MESSAGE = "The project is invalid. Please check the entered value."


class ProjectApi(RestApi):
    def create_project(self, data: ProjectInput, workspace: str | int) -> ProjectOutput:
        if isinstance(workspace, int):
            url = f"/metadata/workspace/{workspace}/project"
        elif isinstance(workspace, str):
            parent_workspace = WorkspaceApi(self.auth).get_workspace(workspace)
            url = f"/metadata/workspace/{parent_workspace.id}/project"
        else:
            raise InvalidReferenceError("workspace", workspace)
        try:
            response = self.post(url, data)
            return ProjectOutput(**response)
        except HttpError as e:
            self._httpErrorHandler.handle_post_http_error(e, "project")
        except IndexError:
            raise ValueError(INVALID_PROJECT_MESSAGE) from None

    def delete_project(self, project: str | int, workspace: str | int | None = None):
        project_output = self.get_project(project, workspace)
        url = f"/metadata/project/{project_output.id}"
        try:
            response = self.delete(url)
            return ProjectOutput(**response)
        except HttpError as e:
            self._httpErrorHandler.handle_delete_http_error(e, "project", project)
        except IndexError:
            raise ValueError(INVALID_PROJECT_MESSAGE) from None

    def get_project(self, project: str | int, workspace: str | int | None = None) -> ProjectOutput:
        parent_workspace = None
        if isinstance(project, int):
            url = f"/metadata/project/{project}"
        elif isinstance(project, str):
            if workspace is None:
                raise MissingReferenceError("workspace")
            parent_workspace = WorkspaceApi(self.auth).get_workspace(workspace)
            url = f"/metadata/workspace/{parent_workspace.id}/project/name/{urllib.parse.quote(project, safe='')}"
        else:
            raise InvalidReferenceError("project", project)
        try:
            response = self.get(url)
            project_output = ProjectOutput(**response)
            if parent_workspace:
                project_output._workspace = parent_workspace
            return project_output
        except HttpError as e:
            self._httpErrorHandler.handle_get_http_error(e, "project", project)
        except IndexError:
            raise ValueError(INVALID_PROJECT_MESSAGE) from None

    def update_project(self, data: ProjectInput, project: str | int, workspace: str | int) -> ProjectOutput:
        if isinstance(workspace, int):
            base_url = f"/metadata/workspace/{workspace}/project"
        elif isinstance(workspace, str):
            base_url = f"/metadata/workspace/name/{urllib.parse.quote(workspace, safe='')}/project"
        else:
            raise InvalidReferenceError("workspace", workspace)
        if isinstance(project, int):
            url = f"/metadata/project/{project}"
        elif isinstance(project, str):
            url = f"{base_url}/name/{urllib.parse.quote(project, safe='')}"
        else:
            raise InvalidReferenceError("project", project)
        try:
            response = self.put(url, data)
            return ProjectOutput(**response)
        except HttpError as e:
            self._httpErrorHandler.handle_put_http_error(e, "project", project)
        except IndexError:
            raise ValueError(INVALID_PROJECT_MESSAGE) from None

    def list_projects(
        self,
        workspace: str | int,
        search: str | None = None,
        page_index: int | None = 1,
        page_size: int | None = 20,
    ) -> PagedResponse[ProjectOutput]:
        if isinstance(workspace, int):
            url = f"/metadata/workspace/project?index={page_index}&size={page_size}&workspaceId={workspace}"
        elif isinstance(workspace, str):
            url = f"/metadata/workspace/project?index={page_index}&size={page_size}&workspaceName={workspace}"
        else:
            raise InvalidReferenceError("workspace", workspace)
        if search:
            url = url + f"&search={search}"
        projects = self.get(url)
        return PagedResponse(
            item_cls=ProjectOutput, total=projects["total"], page=projects["page"], items=projects["items"]
        )



from __future__ import annotations

from vectice.models.resource.metadata.base import (
    DatasetSourceType,
    DatasetSourceUsage,
    Metadata,
)


class DBMetadata(Metadata):
    """Class that describes metadata of dataset that comes from a database."""

    def __init__(
        self,
        dbs: list[MetadataDB],
        size: int,
        usage: DatasetSourceUsage | None = None,
        origin: str | None = None,
    ):
        """Initialize a DBMetadata instance.

        Parameters:
            dbs: The list of databases.
            size: The size of the metadata.
            usage: The usage of the metadata.
            origin: The origin of the metadata.
        """
        super().__init__(size=size, type=DatasetSourceType.DB, usage=usage, origin=origin)
        self.dbs = dbs

    def asdict(self) -> dict:
        return {
            "dbs": [db.asdict() for db in self.dbs],
            "size": self.size,
            "filesCount": None,
            "files": [],
            "type": self.type.value,
            "usage": self.usage.value if self.usage else None,
            "origin": self.origin,
        }


class Column:
    """Model a column of a dataset, like a database column."""

    def __init__(
        self,
        name: str,
        data_type: str,
        is_unique: bool | None = None,
        nullable: bool | None = None,
        is_private_key: bool | None = None,
        is_foreign_key: bool | None = None,
    ):
        """Initialize a column.

        Parameters:
            name: The name of the column.
            data_type: The type of the data contained in the column.
            is_unique: If the column uniquely defines a record.
            nullable: If the column can contain null value.
            is_private_key: If the column uniquely defines a record,
                individually or with other columns (can be null).
            is_foreign_key: If the column refers to another one,
                individually or with other columns (cannot be null).
        """
        self.name = name
        self.data_type = data_type
        self.is_unique = is_unique
        self.nullable = nullable
        self.is_private_key = is_private_key
        self.is_foreign_key = is_foreign_key

    def asdict(self) -> dict:
        return {
            "name": self.name,
            "dataType": self.data_type,
            "isUnique": self.is_unique,
            "nullable": self.nullable,
            "isPK": self.is_private_key,
            "isFK": self.is_foreign_key,
        }


class MetadataDB:
    def __init__(self, name: str, columns: list[Column], rows_number: int, size: int | None = None):
        """Initialize a MetadataDB instance.

        Parameters:
            name: The name of the table.
            columns: The columns that compose the table.
            rows_number: The number of row of the table.
            size: The size of the table.
        """
        self.name = name
        self.size = size
        self.rows_number = rows_number
        self.columns = columns

    def asdict(self) -> dict:
        return {
            "name": self.name,
            "size": self.size,
            "rowsNumber": self.rows_number,
            "columns": [column.asdict() for column in self.columns],
        }



from __future__ import annotations

from typing import TYPE_CHECKING

from vectice.models.datasource.datawrapper.data_wrapper import DataWrapper
from vectice.models.resource.metadata import DatasetSourceOrigin, DatasetSourceUsage, DatasetType, File, FilesMetadata
from vectice.utils.deprecation import deprecate

if TYPE_CHECKING:
    from mypy_boto3_s3 import Client
    from mypy_boto3_s3.type_defs import ListObjectsV2OutputTypeDef, ObjectTypeDef


class S3DataWrapper(DataWrapper):
    """Deprecated. Wrap columnar data and its metadata in AWS S3.

    WARNING: **Wrappers are deprecated.**
    Instead, use [`Dataset`][vectice.models.dataset.Dataset]
    and [`S3Resource`][vectice.models.resource.s3_resource.S3Resource].

    Vectice stores metadata -- data about your dataset -- communicated
    with a DataWrapper.  Your actual dataset is not stored by Vectice.

    This DataWrapper wraps data that you have stored in AWS S3.  You
    assign it to a step.

    ```python
    from vectice import DatasetType, S3DataWrapper, connect
    from boto3.session import Session

    s3_session = Session(  # (1)
        aws_access_key_id="...",
        aws_secret_access_key="...",
        region_name="us-east-1",
    )
    s3_client = s3_session.client(service_name="s3")  # (2)

    my_project = connect(...)  # (3)
    my_phase = my_project.phase(...)  # (4)
    my_iter = my_phase.iteration()  # (5)

    my_iter.step_my_data = S3DataWrapper(
        s3_client,
        bucket_name="my_bucket",
        resource_path="my_resource_path",
        name="My origin dataset name",
        type=DatasetType.ORIGIN,
    )
    ```

    1. See [boto3 sessions][boto3.session.Session].
    1. See [boto3 session client][boto3.session.Session.client].
    1. See [connection][vectice.Connection.connect].
    1. See [phases][vectice.models.Phase].
    1. See [iterations][vectice.models.Iteration].

    Note that these three concepts are distinct, even if easily conflated:

    * Where the data is stored
    * The format at rest (in storage)
    * The format when loaded in a running Python program

    Notably, the statistics collectors provided by Vectice operate
    only on this last and only in the event that the data is loaded as
    a pandas dataframe.
    """

    @deprecate(
        parameter="inputs",
        warn_at="23.1",
        fail_at="23.2",
        remove_at="23.3",
        reason="The 'inputs' parameter is renamed 'derived_from'. "
        "Using 'inputs' will raise an error in v{fail_at}. "
        "The parameter will be removed in v{remove_at}.",
    )
    def __init__(
        self,
        s3_client: Client,
        bucket_name: str,
        resource_path: str,
        name: str,
        usage: DatasetSourceUsage | None = None,
        derived_from: list[int] | None = None,
        inputs: list[int] | None = None,
        type: DatasetType | None = None,
    ):
        """Initialize an S3 data wrapper.

        Parameters:
            s3_client: The client used to interact with Amazon s3.
            bucket_name: The name of the bucket to get data from.
            resource_path: The paths of the resources to get.
            name: The name of the DataWrapper (local to Vectice).
            usage: The usage of the dataset.
            derived_from: The list of dataset ids to create a new dataset from.
            inputs: Deprecated. Use `derived_from` instead.
            type: The type of the dataset.
        """
        if not derived_from and inputs:
            derived_from = inputs

        self.s3_client = s3_client
        self.bucket_name = bucket_name
        self.resource_path = resource_path
        super().__init__(name=name, type=type, usage=usage, derived_from=derived_from)

    def _fetch_data(self) -> dict[str, bytes]:
        s3_objects_list = self._get_s3_objects_list()
        datas = {}
        for s3_object in s3_objects_list["Contents"]:
            object_path = s3_object["Key"]
            data = self._get_aws_data(object_path)
            datas[f"{self.bucket_name}/{object_path}"] = data
        return datas

    def _get_aws_data(self, object_path: str):
        data_response = self.s3_client.get_object(Bucket=self.bucket_name, Key=object_path)
        return data_response["Body"].read()

    def _build_metadata(self) -> FilesMetadata:
        s3_objects_list = self._get_s3_objects_list()
        metadata = self._build_metadata_from_objects(s3_objects_list)
        return metadata

    def _get_s3_objects_list(self) -> ListObjectsV2OutputTypeDef:
        return self.s3_client.list_objects_v2(Bucket=self.bucket_name, Prefix=self.resource_path)

    def _build_metadata_from_objects(self, s3_objects_list: ListObjectsV2OutputTypeDef) -> FilesMetadata:
        if s3_objects_list["KeyCount"] == 0:
            raise NoSuchS3ResourceError(self.bucket_name, self.resource_path)
        s3_objects = s3_objects_list["Contents"]
        files, total_size = self._build_files_list_with_size(s3_objects)
        metadata = FilesMetadata(files=files, origin=DatasetSourceOrigin.S3.value, size=total_size, usage=self.usage)
        return metadata

    def _build_files_list_with_size(self, s3_objects: list[ObjectTypeDef]) -> tuple[list[File], int]:
        files: list[File] = []
        total_size = 0
        for object in s3_objects:
            if not self._is_s3_object_a_folder(object):
                name = object["Key"]
                size = object["Size"]
                file = File(
                    name=name,
                    size=size,
                    fingerprint=self._get_formatted_etag_from_object(object),
                    updated_date=object["LastModified"].isoformat(),
                    created_date=None,
                    uri=f"s3://{self.bucket_name}/{name}",
                )
                total_size += size
                files.append(file)
        return files, total_size

    @staticmethod
    def _get_formatted_etag_from_object(object: ObjectTypeDef) -> str:
        return str(object["ETag"].replace("'", "").replace('"', ""))

    @staticmethod
    def _is_s3_object_a_folder(object: ObjectTypeDef) -> bool:
        return bool(object["Key"].endswith("/"))


class NoSuchS3ResourceError(Exception):
    def __init__(self, bucket: str, resource: str):
        self.message = f"{resource} does not exist in the AWS s3 bucket {bucket}."
        super().__init__(self.message)

    def __str__(self):
        return self.message



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from logging import Logger

    from vectice.api import Client
    from vectice.api.json.last_assets import ActivityTargetType


def _get_last_assets(target_types: list[ActivityTargetType], client: Client, _logger: Logger):
    target_types_array = [target.value for target in target_types]
    try:
        last_asset = client.get_last_assets(target_types_array, {"index": 1, "size": 1}).list[0]
        return last_asset
    except IndexError as e:
        _logger.debug(f"There were no assets with activity found. Sanity check {e}")
        return


def _get_last_user_and_default_workspace(client: Client) -> tuple[str, int | None]:
    asset = client.get_user_and_default_workspace()
    workspace_id = int(asset["defaultWorkspace"]["id"]) if asset["defaultWorkspace"] else None
    return asset["user"]["name"], workspace_id


def _connection_logging(_logger: Logger, user_name: str, host: str, workspace_id: int | None):
    from vectice.utils.logging_utils import CONNECTION_LOGGING

    if workspace_id:
        CONNECTION_LOGGING += r"/workspace/dashboard?w={workspace_id}"  # noqa: N806
        _logger.info(CONNECTION_LOGGING.format(user=user_name, url=host, workspace_id=workspace_id))
        return
    CONNECTION_LOGGING += r"/workspaces"  # noqa: N806
    _logger.info(CONNECTION_LOGGING.format(user=user_name, url=host))



from __future__ import annotations

from typing import TYPE_CHECKING

from gql import gql

from vectice.api.gql_api import GqlApi, Parser

if TYPE_CHECKING:
    from vectice.api.json.model_register import ModelRegisterInput, ModelRegisterOutput

_RETURNS = """
            modelVersion{
                          id
                          name
                          version
                          description
                          algorithmName
                          framework
                          modelId
                          model {
                            name
                            projectId
                          }
            }
            useExistingModel
            __typename
            """


class GqlModelApi(GqlApi):
    def register_model(
        self,
        data: ModelRegisterInput,
        project_id: int,
        phase_id: int | None = None,
        iteration_id: int | None = None,
    ) -> ModelRegisterOutput:
        variables = {"projectId": project_id, "data": data}
        kw = "projectId:$projectId,data:$data"
        variable_types = "$projectId:Float!,$data:ModelRegisterInput!"
        if phase_id:
            variable_types += ",$phaseId:Float!"
            kw += ",phaseId:$phaseId"
            variables["phaseId"] = phase_id
        if iteration_id:
            variable_types += ",$iterationId:Float!"
            kw += ",iterationId:$iterationId"
            variables["iterationId"] = iteration_id
        query_name = "registerModel"
        query = GqlApi.build_query(
            gql_query=query_name,
            variable_types=variable_types,
            returns=_RETURNS,
            keyword_arguments=kw,
            query=False,
        )
        query_built = gql(query)
        response = self.execute(query_built, variables)
        model_output: ModelRegisterOutput = Parser().parse_item(response[query_name])
        return model_output



from __future__ import annotations

from typing import TYPE_CHECKING

from vectice.api._utils import read_nodejs_date
from vectice.api.json.workspace import WorkspaceOutput

if TYPE_CHECKING:
    from datetime import datetime


class ProjectInput(dict):
    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def description(self) -> str:
        return str(self["description"])


class ProjectOutput(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if "workspace" in self:
            self._workspace: WorkspaceOutput = WorkspaceOutput(**self["workspace"])

    def items(self):
        result = []
        for key in self:
            if self[key] is not None:
                result.append((key, self[key]))
        return result

    @property
    def workspace_id(self) -> int:
        return int(self["workspaceId"])

    @property
    def id(self) -> int:
        return int(self["id"])

    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def description(self) -> str | None:
        if "description" in self and self["description"] is not None:
            return str(self["description"])
        else:
            return None

    @property
    def created_date(self) -> datetime | None:
        return read_nodejs_date(str(self["createdDate"]))

    @property
    def updated_date(self) -> datetime | None:
        return read_nodejs_date(str(self["updatedDate"]))

    @property
    def deleted_date(self) -> datetime | None:
        return read_nodejs_date(str(self["deletedDate"]))

    @property
    def version(self) -> int:
        return int(self["version"])

    @property
    def workspace(self) -> WorkspaceOutput:
        return self._workspace



from __future__ import annotations

from vectice.models.resource.metadata import FilesMetadata, MetadataDB
from vectice.utils.deprecation import deprecate


class DatasetRegisterOutput(dict):
    @property
    def version(self) -> str:
        return str(self["version"])

    @property
    def use_existing_version(self) -> bool:
        return bool(self["useExistingVersion"])

    @property
    def use_existing_dataset(self) -> bool:
        return bool(self["useExistingDataset"])

    # TODO: complete the jobrun property
    @property
    def job_run(self) -> str:
        return str(self["jobRun"])


class DatasetRegisterInput(dict):
    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def type(self) -> str:
        return str(self["type"])

    @property
    def data_sources(self) -> list[dict]:
        return list(self["datasetSources"])

    @property
    @deprecate(
        warn_at="23.1",
        fail_at="23.2",
        remove_at="23.3",
        reason="The 'inputs' property is renamed 'derived_from'. "
        "Using 'inputs' will raise an error in v{fail_at}. "
        "The property will be removed in v{remove_at}.",
    )
    def inputs(self) -> list[int]:
        return self.derived_from

    @property
    def derived_from(self) -> list[int]:
        return list(self["inputs"])


class DatasetSourceInput(dict):
    @property
    def type(self) -> str:
        return str(self["type"])

    @property
    def usage(self) -> str:
        return str(self["usage"])

    @property
    def origin(self) -> str:
        return str(self["origin"])

    @property
    def size(self) -> int:
        return int(self["size"])

    @property
    def dbs(self) -> list[MetadataDB]:
        return [MetadataDB(**item) for item in self["dbs"]]

    @property
    def files(self) -> FilesMetadata:
        return FilesMetadata(**self["files"])



from __future__ import annotations

import urllib.parse
from typing import TYPE_CHECKING

from vectice.api.http_error_handlers import InvalidReferenceError
from vectice.api.json import PagedResponse, WorkspaceOutput
from vectice.api.rest_api import HttpError, RestApi

if TYPE_CHECKING:
    from vectice.api.json.workspace import WorkspaceInput


INVALID_WORKSPACE_MESSAGE = "The workspace is invalid. Please check the entered value."


class WorkspaceApi(RestApi):
    def get_workspace(self, workspace: str | int) -> WorkspaceOutput:
        if isinstance(workspace, int):
            url = f"/metadata/workspace/{workspace}"
        elif isinstance(workspace, str):
            url = f"/metadata/workspace/name/{urllib.parse.quote(workspace, safe='')}"
        else:
            raise InvalidReferenceError("workspace", workspace)
        try:
            response = self.get(url)
            return WorkspaceOutput(**response)
        except HttpError as e:
            self._httpErrorHandler.handle_get_http_error(e, "workspace", workspace)
        except TypeError as error:
            raise ValueError(INVALID_WORKSPACE_MESSAGE) from error

    def list_workspaces(
        self, search: str | None = None, page_index: int = 1, page_size: int = 20
    ) -> PagedResponse[WorkspaceOutput]:
        if search is None or search == "":
            url = f"/metadata/workspace?index={page_index}&size={page_size}"
        else:
            url = f"/metadata/workspace?index={page_index}&size={page_size}&search={search}"
        workspaces = self.get(url)
        return PagedResponse(
            item_cls=WorkspaceOutput, total=workspaces["total"], page=workspaces["page"], items=workspaces["items"]
        )

    def create_workspace(self, data: WorkspaceInput) -> WorkspaceOutput:
        url = "/metadata/workspace"
        try:
            response = self.post(url, data.__dict__)
            return WorkspaceOutput(**response)
        except HttpError as e:
            self._httpErrorHandler.handle_post_http_error(e, "workspace")
        except TypeError as error:
            raise ValueError(INVALID_WORKSPACE_MESSAGE) from error

    def update_workspace(self, data: WorkspaceInput, workspace: str | int) -> WorkspaceOutput:
        if isinstance(workspace, int):
            url = f"/metadata/workspace/{workspace}"
        elif isinstance(workspace, str):
            workspace_object = self.get_workspace(workspace)
            url = f"/metadata/workspace/{workspace_object.id}"
        else:
            raise InvalidReferenceError("workspace", workspace)
        try:
            response = self.put(url, data)
            return WorkspaceOutput(**response)
        except HttpError as e:
            self._httpErrorHandler.handle_put_http_error(e, "workspace", workspace)
        except TypeError as error:
            raise ValueError(INVALID_WORKSPACE_MESSAGE) from error



from __future__ import annotations

import logging
from textwrap import dedent
from typing import TYPE_CHECKING

from rich.table import Table

from vectice.api.json.phase import PhaseStatus
from vectice.models.iteration import Iteration
from vectice.models.resource import Resource
from vectice.utils.common_utils import _temp_print
from vectice.utils.deprecation import DeprecationError
from vectice.utils.last_assets import _get_last_user_and_default_workspace
from vectice.utils.logging_utils import format_description

if TYPE_CHECKING:
    from vectice import Connection
    from vectice.api import Client
    from vectice.models import Project, Workspace


_logger = logging.getLogger(__name__)


class Phase:
    """Represent a Vectice phase.

    Phases reflect the real-life phases of the project lifecycle
    (i.e., Business Understanding, Data Preparation, Modeling,
    Deployment, etc.).  The Vectice admin creates the phases of a
    project.

    Phases let you document the goals, assets, and outcomes along with
    the status to organize the project, enforce best practices, allow
    consistency, and capture knowledge.

    Phases contain definitions of steps that are performed
    by data-scientists in order to complete iterations.

    ```tree
    phase 1
        step definition 1
        step definition 2
        step definition 3
    ```

    To get a project's phase:

    ```python
    my_phase = my_project.phase("Business Understanding")
    ```

    Iterations can then be created for this phase,
    to complete the phase steps:

    ```python
    my_origin_dataset = ...
    my_iteration = my_phase.iteration()
    my_iteration.step_origin_dataset = my_origin_dataset
    ```

    NOTE: **Phases and Steps Definitions are created in the Vectice App,
    Iterations are created from the Vectice Python API.**

    See the documentation of [Iterations][vectice.models.Iteration]
    for more information about iterations.
    """

    __slots__ = [
        "_id",
        "_project",
        "_name",
        "_index",
        "_status",
        "_client",
        "_current_iteration",
        "_pointers",
    ]

    def __init__(
        self,
        id: int,
        project: Project,
        name: str,
        index: int,
        status: PhaseStatus = PhaseStatus.NotStarted,
    ):
        """Initialize a phase.

        Vectice users shouldn't need to instantiate Phases manually,
        but here are the phase parameters.

        Parameters:
            id: The phase identifier.
            project: The project to which the phase belongs.
            name: The name of the phase.
            index: The index of the phase.
            status: The status of the phase.
        """
        self._id = id
        self._project = project
        self._name = name
        self._index = index
        self._status = status
        self._client: Client = self._project._client
        self._current_iteration: Iteration | None = None

    def __repr__(self):
        return f"Phase (name='{self.name}', id={self.id}, status='{self.status.name}')"

    def __eq__(self, other: object):
        if not isinstance(other, Phase):
            return NotImplemented
        return self.id == other.id

    @property
    def id(self) -> int:
        """The phase's id.

        Returns:
            The phase's id.
        """
        return self._id

    @id.setter
    def id(self, phase_id: int):
        """Set the phase's id.

        Parameters:
            phase_id: The phase id to set.
        """
        self._id = phase_id

    @property
    def name(self) -> str:
        """The phase's name.

        Returns:
            The phase's name.
        """
        return self._name

    @property
    def index(self) -> int:
        """The phase's index.

        Returns:
            The phase's index.
        """
        return self._index

    @property
    def status(self) -> PhaseStatus:
        """The phase's status.

        Returns:
            The phase's status.
        """
        return self._status

    @property
    def properties(self) -> dict:
        """The phase's name, id, and index.

        Returns:
            A dictionary containing the `name`, `id`, and `index` items.
        """
        return {"name": self.name, "id": self.id, "index": self.index}

    @property
    def iterations(self) -> list[Iteration]:
        """The phase's iterations.

        Returns:
            The phase's iterations.
        """
        iteration_outputs = self._client.list_iterations(self.id)
        return sorted(
            [Iteration(item.id, item.index, self, item.status) for item in iteration_outputs], key=lambda x: x.index
        )

    @property
    def clean_dataset(self) -> None:
        """Deprecated. Use steps' [`clean_dataset`][vectice.models.Step.clean_dataset] instead.

        Raises:
            DeprecationError: Always.
        """
        raise DeprecationError("Clean datasets are now set in steps.")

    @clean_dataset.setter
    def clean_dataset(self, data_source: Resource) -> None:
        """Deprecated. Use steps' [`clean_dataset`][vectice.models.Step.clean_dataset] instead.

        Parameters:
            data_source: The cleaned dataset.

        Raises:
            DeprecationError: Always.
        """
        raise DeprecationError("Clean datasets are now set in steps.")

    def list_step_definitions(self) -> None:
        """List the step definitions to which this phase has access and print the step definitions in a tabular format.

        Returns:
            None
        """
        steps_output = self._client.list_step_definitions(self.id)
        user_name, _ = _get_last_user_and_default_workspace(self._client)

        rich_table = Table(expand=True, show_edge=False)

        rich_table.add_column("step def. id", justify="left", no_wrap=True, min_width=5, max_width=5)
        rich_table.add_column("name", justify="left", no_wrap=True, min_width=5, max_width=10)
        rich_table.add_column("description", justify="left", no_wrap=True, min_width=3, max_width=15)
        rich_table.add_column("shortcut", justify="left", no_wrap=True, min_width=3, max_width=20)

        for count, step in enumerate(reversed(steps_output), 1):
            if count > 10:
                break
            rich_table.add_row(str(step.id), f"{count}. {step.name}", format_description(step.description), step.slug)
        description = (
            f"""There are {len(steps_output)} step definitions in the phase {self.name} for user {user_name}."""
        )
        link = dedent(
            f"""
        For quick access to the list of step definitions in the Vectice web app, visit:
        {self._client.auth._API_BASE_URL}/project/phase/steps?w={self.workspace.id}&pageId={self.id}"""
        ).lstrip()
        _temp_print(description)
        _temp_print(table=rich_table)
        _temp_print(link)

    def iteration(self, index: int | None = None) -> Iteration:
        """Get a (possibly new) iteration.

        Fetch and return an iteration by index. If no index is
        provided, return a new iteration.

        Parameters:
            index: The id of an existing iteration.

        Returns:
            An iteration.
        """
        if index:
            iteration_output = self._client.get_iteration_by_index(self.id, index)
        else:
            iteration_output = self._client.create_iteration(self.id)
        _logger.debug(f"Iteration number {iteration_output.index} (id {iteration_output.id}) successfully retrieved.")
        iteration_object = Iteration(
            id=iteration_output.id,
            index=iteration_output.index,
            phase=self,
            status=iteration_output.status,
        )
        self._current_iteration = iteration_object
        return iteration_object

    @property
    def connection(self) -> Connection:
        """The connection to which this phase belongs.

        Returns:
            The connection to which this phase belongs.
        """
        return self._project.connection

    @property
    def workspace(self) -> Workspace:
        """The workspace to which this phase belongs.

        Returns:
            The workspace to which this phase belongs.
        """
        return self._project.workspace

    @property
    def project(self) -> Project:
        """The project to which this phase belongs.

        Returns:
            The project to which this phase belongs.
        """
        return self._project
